#   Copyright 2022 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import collections
import contextlib
import logging
import multiprocessing as mp
import time
import warnings

from collections import Counter
from collections.abc import Callable, Iterator
from dataclasses import asdict, dataclass, field, replace
from enum import Enum, auto
from typing import Any, Literal, Self, TypeAlias

import arviz as az
import numpy as np
import pymc as pm
import pytensor
import pytensor.tensor as pt

from numpy.typing import NDArray
from packaging import version
from pymc import Model
from pymc.backends.arviz import coords_and_dims_for_inferencedata
from pymc.blocking import DictToArrayBijection, RaveledVars
from pymc.initial_point import make_initial_point_fn
from pymc.model import modelcontext
from pymc.model.core import Point
from pymc.progress_bar import CustomProgress, default_progress_theme
from pymc.pytensorf import compile
from pymc.sampling.mcmc import setup_cores_blas_cores
from pymc.sampling.parallel import _cpu_count, _initialize_multiprocessing_context
from pymc.util import (
    RandomSeed,
    _get_seeds_per_chain,
    get_default_varnames,
)
from pytensor.compile.function.types import Function
from pytensor.compile.mode import FAST_COMPILE, Mode
from pytensor.graph import clone_replace, vectorize_graph
from pytensor.tensor import TensorVariable
from pytensor.tensor.optimize import LRUCache1
from rich.console import Console, Group
from rich.padding import Padding
from rich.progress import TextColumn, TimeElapsedColumn
from rich.table import Column, Table
from rich.text import Text
from threadpoolctl import threadpool_limits

from pymc_extras.inference.laplace_approx.idata import add_data_to_inference_data
from pymc_extras.inference.pathfinder.importance_sampling import (
    importance_sampling as _importance_sampling,
)
from pymc_extras.inference.pathfinder.lbfgs import (
    LBFGS,
    LBFGSException,
    LBFGSInitFailed,
    LBFGSStatus,
    _check_lbfgs_curvature_condition,
)

logger = logging.getLogger(__name__)

REGULARISATION_TERM = 1e-8
DEFAULT_LINKER = "cvm_nogc"

SinglePathfinderFn: TypeAlias = Callable[[int], "PathfinderResult"]


def get_jaxified_logp_of_ravel_inputs(model: Model, jacobian: bool = True) -> Callable:
    """
    Get a JAX function that computes the log-probability of a PyMC model with ravelled inputs.

    Parameters
    ----------
    model : Model
        PyMC model to compute log-probability and gradient.
    jacobian : bool, optional
        Whether to include the Jacobian in the log-probability computation, by default True. Setting to False (not recommended) may result in very high values for pareto k.

    Returns
    -------
    Function
        A JAX function that computes the log-probability of a PyMC model with ravelled inputs.
    """

    from pymc.sampling.jax import get_jaxified_graph

    # TODO: JAX: test if we should get jaxified graph of dlogp as well
    new_logprob, new_input = pm.pytensorf.join_nonshared_inputs(
        model.initial_point(), (model.logp(jacobian=jacobian),), model.value_vars, ()
    )

    logp_func_list = get_jaxified_graph([new_input], new_logprob)

    def logp_func(x):
        return logp_func_list(x)[0]

    return logp_func


def get_logp_dlogp_of_ravel_inputs(
    model: Model, jacobian: bool = True, **compile_kwargs
) -> Function:
    """
    Get the log-probability and its gradient for a PyMC model with ravelled inputs.

    Parameters
    ----------
    model : Model
        PyMC model to compute log-probability and gradient.
    jacobian : bool, optional
        Whether to include the Jacobian in the log-probability computation, by default True. Setting to False (not recommended) may result in very high values for pareto k.
    **compile_kwargs : dict
        Additional keyword arguments to pass to the compile function.

    Returns
    -------
    Function
        A compiled PyTensor function that computes the log-probability and its gradient given ravelled inputs.
    """

    (logP, dlogP), inputs = pm.pytensorf.join_nonshared_inputs(
        model.initial_point(),
        [model.logp(jacobian=jacobian), model.dlogp(jacobian=jacobian)],
        model.value_vars,
    )
    logp_dlogp_fn = compile([inputs], (logP, dlogP), **compile_kwargs)
    logp_dlogp_fn.trust_input = True

    return logp_dlogp_fn


def get_neg_logp_dlogp_of_ravel_inputs(
    model: Model, jacobian: bool = True, **compile_kwargs
) -> Function:
    """Compiled function (x,) -> (-logP, -dlogP) for L-BFGS minimization."""
    (logP, dlogP), inputs = pm.pytensorf.join_nonshared_inputs(
        model.initial_point(),
        [model.logp(jacobian=jacobian), model.dlogp(jacobian=jacobian)],
        model.value_vars,
    )
    neg_logp_dlogp_fn = compile([inputs], (-logP, -dlogP), **compile_kwargs)
    neg_logp_dlogp_fn.trust_input = True
    return neg_logp_dlogp_fn


def get_batched_logp_of_ravel_inputs(
    model: Model, jacobian: bool = True, vectorize: bool = False, **compile_kwargs
) -> Function:
    """Get a batched logP function: (B, N) -> (B,) evaluated in one compiled call.

    Parameters
    ----------
    model : Model
        PyMC model.
    jacobian : bool, optional
        Whether to include the Jacobian, by default True.
    vectorize : bool, optional
        If True, use vectorize_graph instead of pytensor.map. Default False.
        vectorize_graph may be faster for small models but may use more memory
        for large parameter counts.
    **compile_kwargs : dict
        Additional keyword arguments to pass to compile.

    Returns
    -------
    Function
        Compiled function taking a (B, N) array and returning (B,) log-probabilities.
    """
    (logP,), single_input = pm.pytensorf.join_nonshared_inputs(
        model.initial_point(),
        [model.logp(jacobian=jacobian)],
        model.value_vars,
    )
    batch_input = pt.matrix("batch_input", dtype="float64")  # (B, N)
    if vectorize:
        batched_logP = vectorize_graph(logP, replace={single_input: batch_input})
    else:
        # pytensor.map loops over rows of batch_input one at a time; this avoids
        # broadcasting large intermediate tensors (faster than vectorize_graph for
        # models with many parameters).
        batched_logP, _ = pytensor.map(
            fn=lambda x_i: clone_replace([logP], replace={single_input: x_i})[0],
            sequences=[batch_input],
        )
    batched_fn = compile([batch_input], batched_logP, **compile_kwargs)
    batched_fn.trust_input = True
    return batched_fn


def convert_flat_trace_to_idata(
    samples: NDArray,
    include_transformed: bool = False,
    postprocessing_backend: Literal["cpu", "gpu"] = "cpu",
    inference_backend: Literal["pymc", "blackjax"] = "pymc",
    model: Model | None = None,
    importance_sampling: Literal["psis", "psir", "identity"] | None = "psis",
) -> az.InferenceData:
    """convert flattened samples to arviz InferenceData format.

    Parameters
    ----------
    samples : NDArray
        flattened samples
    include_transformed : bool
        whether to include transformed variables
    postprocessing_backend : str
        backend for postprocessing transformations, either "cpu" or "gpu"
    inference_backend : str
        backend for inference, either "pymc" or "blackjax"
    model : Model | None
        pymc model for variable transformations
    importance_sampling : str
        importance sampling method used, affects input samples shape

    Returns
    -------
    InferenceData
        arviz inference data object
    """

    if importance_sampling is None:
        # samples.ndim == 3 in this case, otherwise ndim == 2
        num_paths, num_pdraws, N = samples.shape
        samples = samples.reshape(-1, N)

    model = modelcontext(model)
    ip = model.initial_point()
    ip_point_map_info = DictToArrayBijection.map(ip).point_map_info
    trace = collections.defaultdict(list)
    for sample in samples:
        raveld_vars = RaveledVars(sample, ip_point_map_info)
        point = DictToArrayBijection.rmap(raveld_vars, ip)
        for p, v in point.items():
            trace[p].append(v.tolist())

    trace = {k: np.asarray(v)[None, ...] for k, v in trace.items()}

    var_names = model.unobserved_value_vars
    vars_to_sample = list(get_default_varnames(var_names, include_transformed=include_transformed))
    logger.info("Transforming variables...")

    if inference_backend == "pymc":
        # vectorize_graph batches over trace dims; output size matches input, no extra intermediates.
        new_shapes = [v.ndim * (None,) for v in trace.values()]
        replace = {
            var: pt.tensor(dtype="float64", shape=new_shapes[i])
            for i, var in enumerate(model.value_vars)
        }

        outputs = vectorize_graph(vars_to_sample, replace=replace)

        fn = pytensor.function(
            inputs=[*list(replace.values())],
            outputs=outputs,
            mode=FAST_COMPILE,
            on_unused_input="ignore",
        )
        fn.trust_input = True
        result = fn(*list(trace.values()))
        if not isinstance(result, list):
            result = [result]

        if importance_sampling is None:
            result = [res.reshape(num_paths, num_pdraws, *res.shape[2:]) for res in result]

    elif inference_backend == "blackjax":
        import jax

        from pymc.sampling.jax import get_jaxified_graph

        jax_fn = get_jaxified_graph(inputs=model.value_vars, outputs=vars_to_sample)
        result = jax.vmap(jax.vmap(jax_fn))(
            *jax.device_put(list(trace.values()), jax.devices(postprocessing_backend)[0])
        )

    trace = {v.name: r for v, r in zip(vars_to_sample, result)}
    coords, dims = coords_and_dims_for_inferencedata(model)
    idata = az.from_dict(trace, dims=dims, coords=coords)

    return idata


def alpha_recover(
    x: TensorVariable, g: TensorVariable
) -> tuple[TensorVariable, TensorVariable, TensorVariable]:
    """compute the diagonal elements of the inverse Hessian at each iterations of L-BFGS and filter updates.

    Parameters
    ----------
    x : TensorVariable
        position array, shape (L+1, N)
    g : TensorVariable
        gradient array, shape (L+1, N)

    Returns
    -------
    alpha : TensorVariable
        diagonal elements of the inverse Hessian at each iteration of L-BFGS, shape (L, N)
    s : TensorVariable
        position differences, shape (L, N)
    z : TensorVariable
        gradient differences, shape (L, N)

    Notes
    -----
    shapes: L=batch_size, N=num_params
    """

    def compute_alpha_l(s_l, z_l, alpha_lm1) -> TensorVariable:
        # alpha_lm1: (N,)
        # s_l: (N,)
        # z_l: (N,)
        a = pt.sum(alpha_lm1 * z_l**2)
        b = pt.sum(z_l * s_l)
        c = pt.sum(s_l**2 / alpha_lm1)
        inv_alpha_l = (
            a / (b * alpha_lm1)
            + z_l ** 2 / b
            - (a * s_l ** 2) / (b * c * alpha_lm1**2)
        )  # fmt:off
        return 1.0 / inv_alpha_l

    Lp1, N = x.shape
    s = pt.diff(x, axis=0)
    z = pt.diff(g, axis=0)
    alpha_l_init = pt.ones(N)

    alpha = pytensor.scan(
        fn=compute_alpha_l,
        outputs_info=alpha_l_init,
        sequences=[s, z],
        n_steps=Lp1 - 1,
        allow_gc=False,
        return_updates=False,
    )

    # alpha: (L, N)
    return alpha, s, z


def alpha_step_numpy(alpha_prev: NDArray, s: NDArray, z: NDArray) -> NDArray:
    """Pure-numpy single-step alpha update. Stays in sync with compute_alpha_l in alpha_recover.

    Parameters
    ----------
    alpha_prev : NDArray
        previous alpha vector, shape (N,)
    s : NDArray
        position diff x[l] - x[l-1], shape (N,)
    z : NDArray
        gradient diff g[l] - g[l-1], shape (N,)

    Returns
    -------
    NDArray
        updated alpha, shape (N,)
    """
    a = np.sum(alpha_prev * z**2)
    b = np.sum(z * s)
    c = np.sum(s**2 / alpha_prev)
    # Guard: when b≈0 or c≈0 the formula divides by zero; keep alpha_prev to avoid NaN propagation
    z_sq = float(np.sum(z**2)) + 1e-30
    if abs(b) < 1e-14 * z_sq or c <= 0 or not np.isfinite(c):
        return alpha_prev.copy()
    inv_alpha = (
        a / (b * alpha_prev)
        + z**2 / b
        - (a * s**2) / (b * c * alpha_prev**2)
    )  # fmt: off
    alpha_out = 1.0 / inv_alpha
    if not np.all(np.isfinite(alpha_out)) or np.any(alpha_out <= 0):
        return alpha_prev.copy()
    return alpha_out


def _bfgs_sample_pt(
    x: TensorVariable,
    g: TensorVariable,
    alpha: TensorVariable,
    S: TensorVariable,
    Z: TensorVariable,
    u: TensorVariable,
    J: int,
    N: int,
) -> tuple[TensorVariable, TensorVariable, TensorVariable]:
    """Symbolic L-BFGS inverse-Hessian sample.

    The dense vs sparse path is selected at graph-construction time from
    the compile-time constants N and J, so only one branch is ever compiled.

    Parameters
    ----------
    x : (N,) position
    g : (N,) gradient
    alpha : (N,) diagonal scaling
    S : (N, J) position-diff ring buffer
    Z : (N, J) gradient-diff ring buffer
    u : (M, N) standard-normal draws — M is a dynamic runtime dimension
    J : int, L-BFGS history size (compile-time constant)
    N : int, number of parameters (compile-time constant)

    Returns
    -------
    phi : (M, N) samples
    logQ : (M,) log-density under the approximation
    inv_hessian_diag : (N,) diagonal of sampler covariance used for `phi`
    """
    J2 = 2 * J

    # Inverse Hessian factors
    E = pt.triu(S.T @ Z) + pt.eye(J) * REGULARISATION_TERM  # (J, J)
    eta = pt.diag(E)  # (J,)
    AZ = alpha[:, None] * Z  # (N, J)
    beta = pt.concatenate([AZ, S], axis=1)  # (N, 2J)
    E_inv = pt.linalg.solve_triangular(E, pt.eye(J), lower=False)  # (J, J)
    block_dd = E_inv.T @ (pt.diag(eta) + Z.T @ AZ) @ E_inv  # (J, J)
    top = pt.concatenate([pt.zeros((J, J)), -E_inv], axis=1)  # (J, 2J)
    bot = pt.concatenate([-E_inv.T, block_dd], axis=1)  # (J, 2J)
    gamma = pt.concatenate([top, bot], axis=0)  # (2J, 2J)

    inv_sqrt_alpha = 1.0 / pt.sqrt(alpha)
    sqrt_alpha = pt.sqrt(alpha)

    if J2 >= N:
        # Dense path: form H_inv explicitly (small-N regime, O(N²) is fine)
        isa_d = pt.diag(inv_sqrt_alpha)  # (N, N)
        sa_d = pt.diag(sqrt_alpha)  # (N, N)
        IdN = pt.eye(N) * (1.0 + REGULARISATION_TERM)
        H_inv = sa_d @ (IdN + isa_d @ beta @ gamma @ beta.T @ isa_d) @ sa_d
        Lchol = pt.linalg.cholesky(H_inv, lower=False)  # (N, N)
        logdet = 2.0 * pt.sum(pt.log(pt.abs(pt.diag(Lchol))))
        mu = x - H_inv @ g
        phi = (mu[:, None] + Lchol @ u.T).T  # (M, N)
        inv_hessian_diag = pt.diag(H_inv)
    else:
        # Sparse path: economy QR avoids O(N²) matrices (large-N regime)
        Q, R = pt.linalg.qr(beta * inv_sqrt_alpha[:, None], mode="reduced")  # Q:(N,2J), R:(2J,2J)
        I2J = pt.eye(J2) * (1.0 + REGULARISATION_TERM)
        Lchol = pt.linalg.cholesky(I2J + R @ gamma @ R.T, lower=False)  # (2J, 2J)
        logdet = 2.0 * pt.sum(pt.log(pt.abs(pt.diag(Lchol)))) + pt.sum(pt.log(alpha))
        btg = beta.T @ g[:, None]  # (2J, 1)
        mu = x - ((alpha * g)[:, None] + beta @ (gamma @ btg))[:, 0]  # (N,)
        QtU = Q.T @ u.T  # (2J, M)
        phi = (mu[:, None] + sqrt_alpha[:, None] * (Q @ ((Lchol - pt.eye(J2)) @ QtU) + u.T)).T
        # diag(Q (L L^T - I) Q^T) = row_norm_sq(Q L) - row_norm_sq(Q)
        ql = Q @ Lchol
        inv_hessian_diag = alpha * (1.0 + pt.sum(ql * ql, axis=1) - pt.sum(Q * Q, axis=1))

    logQ = -0.5 * (logdet + pt.sum(u * u, axis=-1) + N * np.log(2.0 * np.pi))
    return phi, logQ, inv_hessian_diag


def make_pathfinder_sample_fn(
    model: Model,
    N: int,
    J: int,
    jacobian: bool,
    compile_kwargs: dict,
    vectorize: bool = False,
    return_inv_hessian_diag: bool = False,
) -> Function:
    """Compile a single PyTensor function covering bfgs sample + batched logP evaluation.

    S and Z are passed as explicit inputs (not shared variables) so each caller
    can pass its own arrays, avoiding shared mutable state.

    The number of draws M is a dynamic runtime dimension — the same compiled
    function handles both ELBO estimation (small M) and final sampling (large M).

    Parameters
    ----------
    model : Model
    N : int, number of unconstrained parameters
    J : int, L-BFGS history size (maxcor)
    jacobian : bool
    compile_kwargs : dict
    vectorize : bool, optional
        If True, use vectorize_graph instead of pytensor.map. Default False.
    return_inv_hessian_diag : bool, optional
        If True, include inverse-Hessian diagonal as 4th output.

    Returns
    -------
    fn : Function
        Compiled: (x, g, alpha, s_win, z_win, u) → (phi, logQ, logP[, inv_hessian_diag])
        where s_win, z_win are (N, J), u is (M, N), and M is a dynamic dimension.
    """
    (logP_single,), single_input = pm.pytensorf.join_nonshared_inputs(
        model.initial_point(),
        [model.logp(jacobian=jacobian)],
        model.value_vars,
    )

    x_sym = pt.vector("x", dtype="float64")
    g_sym = pt.vector("g", dtype="float64")
    alpha_sym = pt.vector("alpha", dtype="float64")
    s_win_sym = pt.matrix("s_win", dtype="float64")  # (N, J)
    z_win_sym = pt.matrix("z_win", dtype="float64")  # (N, J)
    u_sym = pt.matrix("u", dtype="float64")  # (M, N) — M is dynamic

    phi_sym, logQ_sym, inv_hessian_diag_sym = _bfgs_sample_pt(
        x_sym, g_sym, alpha_sym, s_win_sym, z_win_sym, u_sym, J, N
    )

    if vectorize:
        batched_logP_sym = vectorize_graph(logP_single, replace={single_input: phi_sym})
    else:
        batched_logP_sym = pytensor.map(
            fn=lambda x_i: clone_replace([logP_single], replace={single_input: x_i})[0],
            sequences=[phi_sym],
            return_updates=False,
        )

    outputs = [phi_sym, logQ_sym, batched_logP_sym]
    if return_inv_hessian_diag:
        outputs.append(inv_hessian_diag_sym)

    fn = pytensor.function(
        [
            pytensor.In(x_sym, borrow=True),
            pytensor.In(g_sym, borrow=True),
            pytensor.In(alpha_sym, borrow=True),
            pytensor.In(s_win_sym, borrow=True),
            pytensor.In(z_win_sym, borrow=True),
            pytensor.In(u_sym, borrow=True),
        ],
        outputs,
        **compile_kwargs,
    )
    fn.trust_input = True
    return fn


class PathStatus(Enum):
    """
    Statuses of a single-path pathfinder.
    """

    SUCCESS = auto()
    ELBO_ARGMAX_AT_ZERO = auto()
    # Statuses that lead to Exceptions:
    INVALID_LOGP = auto()
    INVALID_LOGQ = auto()
    LBFGS_FAILED = auto()
    PATH_FAILED = auto()
    SINGLE_STEP = auto()


FAILED_PATH_STATUS = [
    PathStatus.INVALID_LOGP,
    PathStatus.INVALID_LOGQ,
    PathStatus.LBFGS_FAILED,
    PathStatus.PATH_FAILED,
    PathStatus.SINGLE_STEP,
]


class PathException(Exception):
    """
    raises a PathException if the path failed.
    """

    DEFAULT_MESSAGE = "Path failed."

    def __init__(self, message=None, status: PathStatus = PathStatus.PATH_FAILED) -> None:
        super().__init__(message or self.DEFAULT_MESSAGE)
        self.status = status


class PathInvalidLogP(PathException):
    """
    raises a PathException if all the logP values in a path are not finite.
    """

    DEFAULT_MESSAGE = "Path failed because all the logP values in a path are not finite."

    def __init__(self, message=None) -> None:
        super().__init__(message or self.DEFAULT_MESSAGE, PathStatus.INVALID_LOGP)


class PathInvalidLogQ(PathException):
    """
    raises a PathException if all the logQ values in a path are not finite.
    """

    DEFAULT_MESSAGE = "Path failed because all the logQ values in a path are not finite."

    def __init__(self, message=None) -> None:
        super().__init__(message or self.DEFAULT_MESSAGE, PathStatus.INVALID_LOGQ)


class SingleStepPathException(PathException):
    """
    raises when the path has only one LBFGS step (insufficient for valid approximation).
    """

    DEFAULT_MESSAGE = "Path failed because only a single step was performed."

    def __init__(self, message=None) -> None:
        super().__init__(message or self.DEFAULT_MESSAGE, PathStatus.SINGLE_STEP)


class LBFGSStreamingCallback:
    """Streaming LBFGS callback: computes ELBO at each accepted step, O(J*N + M*N) peak memory.

    Instead of collecting the full (L+1, N) history, it processes each accepted step
    immediately and tracks only the best state seen so far.

    Parameters
    ----------
    value_grad_fn : Callable
        Single-entry cached value/gradient function (e.g. LRUCache1 from pytensor.tensor.optimize).
    x0 : NDArray
        Initial position, shape (N,).
    sample_logp_fn : Callable
        Compiled PyTensor function (x, g, alpha, s_win, z_win, u) → (phi, logQ, logP).
        Built by make_pathfinder_sample_fn.
    num_elbo_draws : int
        Number of draws per step for ELBO estimation.
    rng : np.random.Generator
        Random number generator for draw generation.
    J : int
        L-BFGS history size (maxcor).
    epsilon : float
        Tolerance for the LBFGS update condition.
    progress_callback : Callable | None
        Optional progress reporting.
    on_step_callback : Callable | None
        If set, called after each accepted step with (x, g, alpha, s_win, z_win, elbo).
        Used by fixture generation to record per-step state.
    """

    def __init__(
        self,
        value_grad_fn: Callable,
        x0: NDArray,
        sample_logp_fn: Callable,
        num_elbo_draws: int,
        rng: np.random.Generator,
        J: int,
        epsilon: float,
        progress_callback: Callable | None = None,
        on_step_callback: Callable | None = None,
    ) -> None:
        self.value_grad_fn = value_grad_fn
        self.sample_logp_fn = sample_logp_fn
        self.num_elbo_draws = num_elbo_draws
        self._rng = rng
        self.J = J
        self.epsilon = epsilon
        self.progress_callback = progress_callback
        self.on_step_callback = on_step_callback

        N = x0.shape[0]
        self._N = N
        _, g0 = value_grad_fn(x0)

        self.x_prev: NDArray = x0.copy()
        self.g_prev: NDArray = np.array(g0, dtype=np.float64)
        self.alpha_prev: NDArray = np.ones(N, dtype=np.float64)

        # Ring buffer: numpy arrays passed as inputs to sample_logp_fn each call.
        # Thread-safe: no shared mutable state across concurrent invocations.
        self.s_win: NDArray = np.zeros((N, J), dtype=np.float64)
        self.z_win: NDArray = np.zeros((N, J), dtype=np.float64)
        self.win_idx: int = -1
        self.best_elbo: float = -np.inf
        self.best_state: dict = {}
        self.best_step_idx: int = 0
        self.step_count: int = 0
        self.any_valid: bool = False
        self.current_elbo: float | None = None
        self._start_time: float = time.time()

    def __call__(self, x: NDArray) -> float | None:
        """Process one accepted LBFGS step. Returns current_elbo for testability."""
        value, g = self.value_grad_fn(x)

        s = x - self.x_prev
        z = g - self.g_prev

        if not (np.all(np.isfinite(g)) and np.isfinite(value)):
            self.current_elbo = None
            return None
        if not _check_lbfgs_curvature_condition(s, z, self.epsilon):
            self.current_elbo = None
            return None

        alpha = alpha_step_numpy(self.alpha_prev, s, z)

        # Ring-buffer update (numpy, O(N))
        self.win_idx = (self.win_idx + 1) % self.J
        self.s_win[:, self.win_idx] = s
        self.z_win[:, self.win_idx] = z

        # Sample + logP in a single compiled call. Pass s_win/z_win as inputs.
        u = self._rng.standard_normal((self.num_elbo_draws, self._N))
        try:
            sample_out = self.sample_logp_fn(x, g, alpha, self.s_win, self.z_win, u)
            _, logQ, logP = sample_out[:3]
            logP = np.asarray(logP)
            logQ = np.asarray(logQ)
            finite = np.isfinite(logP)
            if not np.any(finite):
                elbo = -np.inf
            else:
                logP_safe = np.where(finite, logP, -np.inf)
                elbo = float(np.mean(logP_safe - logQ))
                if not np.isfinite(elbo):
                    elbo = -np.inf
        except Exception:
            elbo = -np.inf

        if np.isfinite(elbo):
            self.any_valid = True

        if self.on_step_callback is not None:
            self.on_step_callback(x, g, alpha, self.s_win.copy(), self.z_win.copy(), elbo)

        if elbo > self.best_elbo:
            self.best_elbo = elbo
            self.best_state = {
                "alpha": alpha.copy(),
                "s_win": self.s_win.copy(),
                "z_win": self.z_win.copy(),
                "win_idx": self.win_idx,
                "x": x.copy(),
                "g": g.copy(),
            }
            self.best_step_idx = self.step_count

        self.alpha_prev = alpha
        self.x_prev = x.copy()
        self.g_prev = g.copy()
        self.step_count += 1

        self.current_elbo = elbo if np.isfinite(elbo) else None

        if self.progress_callback is not None:
            best_elbo = self.best_elbo if np.isfinite(self.best_elbo) else None
            current_elbo = self.current_elbo
            elapsed = time.time() - self._start_time
            steps_per_sec = self.step_count / elapsed if elapsed > 0 else None
            step_size = float(np.linalg.norm(s))
            self.progress_callback(
                {
                    "lbfgs_steps": self.step_count,
                    "best_elbo": best_elbo,
                    "best_ind": self.best_step_idx,
                    "current_elbo": current_elbo,
                    "step_size": step_size,
                    "steps_per_sec": steps_per_sec,
                }
            )

        return elbo


def make_single_pathfinder_fn(
    model,
    num_draws: int,
    maxcor: int | None,
    maxiter: int,
    ftol: float,
    gtol: float,
    maxls: int,
    num_elbo_draws: int,
    jitter: float,
    epsilon: float,
    max_init_retries: int = 10,
    pathfinder_kwargs: dict[str, Any] = {},
    compile_kwargs: dict[str, Any] = {},
) -> SinglePathfinderFn:
    """
    returns a seedable single-path pathfinder function, where it executes a compiled function that performs the local approximation and sampling part of the Pathfinder algorithm.

    Parameters
    ----------
    model : pymc.Model
        The PyMC model to fit the Pathfinder algorithm to.
    num_draws : int
        Number of samples to draw from the single-path approximation.
    maxcor : int | None
        Maximum number of iterations for the L-BFGS optimisation.
    maxiter : int
        Maximum number of iterations for the L-BFGS optimisation.
    ftol : float
        Tolerance for the decrease in the objective function.
    gtol : float
        Tolerance for the norm of the gradient.
    maxls : int
        Maximum number of line search steps for the L-BFGS algorithm.
    num_elbo_draws : int
        Number of draws for the Evidence Lower Bound (ELBO) estimation.
    jitter : float
        Amount of jitter to apply to initial points. Note that Pathfinder may be highly sensitive to the jitter value. It is recommended to increase num_paths when increasing the jitter value.
    epsilon : float
        value used to filter out large changes in the direction of the update gradient at each iteration l in L. Iteration l is only accepted if s·z >= epsilon * ||z||² for each l in L. Matches Zhang et al. (2022) Algorithm 3 with default epsilon=1e-12.
    max_init_retries : int, optional
        Maximum number of re-jitter retries when LBFGSInitFailed is raised (i.e. the initial
        point yields non-finite value/gradient or the first step is rejected). Each retry uses
        a different jitter seed. Defaults to 10.
    pathfinder_kwargs : dict
        Additional keyword arguments for the Pathfinder algorithm. Supported keys:
        ``jacobian`` (bool, default True), ``vectorize`` (bool, default False).
        If ``vectorize`` is True, use vectorize_graph instead of pytensor.map for
        batched logP evaluation; may be faster for small models.
    compile_kwargs : dict
        Additional keyword arguments for the PyTensor compiler. If not provided, a
        performant default is used.

    Returns
    -------
    single_pathfinder_fn : Callable
        A seedable single-path pathfinder function that accepts ``(random_seed, progress_callback=None)``.
    """

    compile_kwargs = {"mode": Mode(linker=DEFAULT_LINKER), **compile_kwargs}
    jacobian = pathfinder_kwargs.get("jacobian", True)
    vectorize = pathfinder_kwargs.get("vectorize", False)
    logp_dlogp_kwargs = {"jacobian": jacobian, **compile_kwargs}

    neg_logp_dlogp_func = get_neg_logp_dlogp_of_ravel_inputs(model, **logp_dlogp_kwargs)

    # initial point
    # TODO: remove make_initial_points function when feature request is implemented: https://github.com/pymc-devs/pymc/issues/7555
    ipfn = make_initial_point_fn(model=model)
    ip = Point(ipfn(None), model=model)
    x_base = DictToArrayBijection.map(ip).data
    N = x_base.shape[0]

    sample_logp_func = make_pathfinder_sample_fn(
        model,
        N,
        maxcor,
        jacobian=jacobian,
        compile_kwargs=compile_kwargs,
        vectorize=vectorize,
    )
    final_sample_logp_func = make_pathfinder_sample_fn(
        model,
        N,
        maxcor,
        jacobian=jacobian,
        compile_kwargs=compile_kwargs,
        vectorize=vectorize,
        return_inv_hessian_diag=True,
    )

    def _check_lbfgs_status(status):
        if status in {LBFGSStatus.INIT_FAILED, LBFGSStatus.INIT_FAILED_LOW_UPDATE_PCT}:
            raise LBFGSInitFailed(status)
        elif status == LBFGSStatus.LBFGS_FAILED:
            raise LBFGSException()

    def _make_result(
        psi, logP_psi, logQ_psi, lbfgs_niter, elbo_argmax, inv_hessian_diag, lbfgs_status
    ):
        if np.all(~np.isfinite(logQ_psi)):
            raise PathInvalidLogQ()
        path_status = PathStatus.ELBO_ARGMAX_AT_ZERO if elbo_argmax == 0 else PathStatus.SUCCESS
        return PathfinderResult(
            samples=psi,
            logP=logP_psi,
            logQ=logQ_psi,
            lbfgs_niter=lbfgs_niter,
            elbo_argmax=elbo_argmax,
            inv_hessian_diag=inv_hessian_diag,
            lbfgs_status=lbfgs_status,
            path_status=path_status,
        )

    def single_pathfinder_fn(
        random_seed: int, progress_callback: Callable | None = None
    ) -> PathfinderResult:
        if progress_callback is not None:
            progress_callback({"status": "running"})

        # Per-path independent copies of compiled functions for process safety.
        local_neg_logp_dlogp = neg_logp_dlogp_func.copy(share_memory=False)
        cached_fn = LRUCache1(local_neg_logp_dlogp, copy_x=True)

        local_lbfgs = LBFGS(cached_fn, maxcor, maxiter, ftol, gtol, maxls, epsilon)

        local_sample_logp = sample_logp_func.copy(share_memory=False)
        local_final_sample_logp = final_sample_logp_func.copy(share_memory=False)

        lbfgs_status = LBFGSStatus.LBFGS_FAILED  # default before LBFGS runs
        try:
            # Derive base seeds once from random_seed. init_seed is an int so we can
            # safely offset it per retry attempt (avoids Generator arithmetic issues).
            _base_init, elbo_seed, final_seed = _get_seeds_per_chain(random_seed, 3)

            for attempt in range(max_init_retries + 1):
                try:
                    init_seed = _base_init + attempt  # different jitter per retry
                    rng = np.random.default_rng(init_seed)
                    jitter_value = rng.uniform(-jitter, jitter, size=x_base.shape)
                    x0 = x_base + jitter_value

                    # Fresh NumPy RNG each attempt → reproducible regardless
                    # of how many ELBO steps the previous attempt took.
                    elbo_rng = np.random.default_rng(elbo_seed + attempt)
                    cached_fn.clear_cache()

                    streaming_cb = LBFGSStreamingCallback(
                        value_grad_fn=cached_fn,
                        x0=x0,
                        sample_logp_fn=local_sample_logp,
                        num_elbo_draws=num_elbo_draws,
                        rng=elbo_rng,
                        J=maxcor,
                        epsilon=epsilon,
                        progress_callback=progress_callback,
                    )

                    lbfgs_niter, lbfgs_status = local_lbfgs.minimize_streaming(streaming_cb, x0)
                    _check_lbfgs_status(lbfgs_status)

                    if lbfgs_niter < 2:
                        raise SingleStepPathException()

                    if not streaming_cb.any_valid:
                        raise PathInvalidLogP()

                    elbo_argmax = streaming_cb.best_step_idx
                    best_state = streaming_cb.best_state

                    final_rng = np.random.default_rng(final_seed)
                    u_final = final_rng.standard_normal((num_draws, N))
                    if progress_callback is not None:
                        progress_callback(
                            {"status": "sampling", "current_elbo": None, "step_size": None}
                        )

                    sample_out = local_final_sample_logp(
                        best_state["x"],
                        best_state["g"],
                        best_state["alpha"],
                        best_state["s_win"],
                        best_state["z_win"],
                        u_final,
                    )
                    phi_final, logQ_psi_flat, logP_psi_flat, inv_hessian_diag = sample_out
                    phi_final = np.asarray(phi_final)
                    logQ_psi_flat = np.asarray(logQ_psi_flat)
                    logP_psi_flat = np.asarray(logP_psi_flat)
                    inv_hessian_diag = np.asarray(inv_hessian_diag)[None]
                    # Add batch dim L=1 to match downstream expectations
                    psi = phi_final[None]  # (1, M, N)
                    logP_psi = logP_psi_flat[None]  # (1, M)
                    logQ_psi = logQ_psi_flat[None]  # (1, M)

                    break  # success — exit retry loop

                except (LBFGSInitFailed, SingleStepPathException) as e:
                    if attempt < max_init_retries:
                        logger.debug(
                            "%s on attempt %d/%d, retrying with different jitter...",
                            type(e).__name__,
                            attempt + 1,
                            max_init_retries,
                        )
                        if progress_callback is not None:
                            progress_callback({"status": f"retry {attempt + 1}"})
                    else:
                        if progress_callback is not None:
                            progress_callback(
                                {
                                    "status": "lbfgs_fail"
                                    if isinstance(e, LBFGSInitFailed)
                                    else "failed"
                                }
                            )
                        if isinstance(e, LBFGSInitFailed):
                            return PathfinderResult(
                                lbfgs_status=e.status,
                                path_status=PathStatus.LBFGS_FAILED,
                            )
                        return PathfinderResult(
                            lbfgs_status=lbfgs_status,
                            path_status=PathStatus.SINGLE_STEP,
                        )

            result = _make_result(
                psi,
                logP_psi,
                logQ_psi,
                lbfgs_niter,
                elbo_argmax,
                inv_hessian_diag,
                lbfgs_status,
            )
            if progress_callback is not None:
                status_str = (
                    "elbo@0" if result.path_status == PathStatus.ELBO_ARGMAX_AT_ZERO else "ok"
                )
                progress_callback(
                    {"status": status_str, "lbfgs_steps": int(lbfgs_niter), "best_ind": elbo_argmax}
                )
            return result

        except LBFGSException as e:
            if progress_callback is not None:
                progress_callback({"status": "lbfgs_fail"})
            return PathfinderResult(
                lbfgs_status=e.status,
                path_status=PathStatus.LBFGS_FAILED,
            )
        except PathException as e:
            if progress_callback is not None:
                progress_callback({"status": "failed"})
            return PathfinderResult(
                lbfgs_status=lbfgs_status,
                path_status=e.status,
            )

    return single_pathfinder_fn


def _default_cores(num_paths: int, cores: int | None) -> int:
    """Default cores for parallel pathfinder, mirroring pm.sample."""
    if cores is not None:
        return min(cores, num_paths)
    return min(4, _cpu_count(), num_paths)


class _PipeCallback:
    """Picklable progress callback that relays updates through a multiprocessing Pipe."""

    def __init__(self, conn: Any, idx: int) -> None:
        self.conn = conn
        self.idx = idx

    def __call__(self, info: dict) -> None:
        try:
            self.conn.send((self.idx, info))
        except Exception:
            pass


def _run_path(
    fn_pickled: bytes,
    seed: int,
    path_id: int,
    progress_conn: Any,
    blas_cores: int | None,
    mp_start_method: str,
) -> "PathfinderResult":
    """Worker: unpickle fn, run with threadpool_limits when blas_cores set (non-fork)."""
    import cloudpickle

    from pytensor.compile.compilelock import lock_ctx

    ctx = (
        threadpool_limits(limits=blas_cores)
        if mp_start_method != "fork" and blas_cores is not None
        else contextlib.nullcontext()
    )
    with ctx:
        with lock_ctx(timeout=-1):
            fn = cloudpickle.loads(fn_pickled)
        cb = _PipeCallback(progress_conn, path_id)
        try:
            return fn(seed, cb)
        finally:
            try:
                progress_conn.send((path_id, None))  # sentinel
            except Exception:
                pass


def _execute_concurrently(
    fn: SinglePathfinderFn,
    seeds: list[int],
    cores: int,
    blas_cores: int | None,
    progress_callbacks: list[Callable | None] | None = None,
    mp_ctx: mp.context.BaseContext | str | None = None,
) -> Iterator["PathfinderResult"]:
    """Execute pathfinder runs concurrently via ProcessPoolExecutor.

    Uses Pipe instead of Manager().Queue() to avoid spawn bootstrapping issues
    when mp_ctx is 'spawn' and the main module is still loading.
    """
    import threading

    from concurrent.futures import ProcessPoolExecutor, as_completed

    import cloudpickle

    mp_ctx = _initialize_multiprocessing_context(mp_ctx, quiet=True)
    fn_pickled = cloudpickle.dumps(fn, protocol=-1)

    # One pipe per worker; avoids Manager() which spawns a process and triggers
    # "An attempt has been made to start a new process before bootstrapping" when
    # the main module is still loading (e.g. script run without if __name__ guard).
    n_workers = len(seeds)
    parent_conns = []
    child_conns = []
    for _ in range(n_workers):
        parent, child = mp_ctx.Pipe(duplex=False)
        parent_conns.append(parent)
        child_conns.append(child)

    sentinel_count: list[int] = [0]
    sentinel_lock = threading.Lock()

    def _listener() -> None:
        from multiprocessing.connection import wait

        while True:
            ready = wait(parent_conns, timeout=0.1)
            for conn in ready:
                try:
                    idx, info = conn.recv()
                except EOFError:
                    continue
                if info is None:
                    with sentinel_lock:
                        sentinel_count[0] += 1
                    if sentinel_count[0] >= n_workers:
                        return
                    continue
                if (
                    progress_callbacks
                    and idx < len(progress_callbacks)
                    and progress_callbacks[idx] is not None
                ):
                    progress_callbacks[idx](info)

    listener = threading.Thread(target=_listener, daemon=True)
    listener.start()

    def _run_executor():
        with ProcessPoolExecutor(max_workers=cores, mp_context=mp_ctx) as executor:
            futures = {
                executor.submit(
                    _run_path,
                    fn_pickled,
                    seed,
                    i,
                    child_conns[i],
                    blas_cores,
                    mp_ctx.get_start_method(),
                ): i
                for i, seed in enumerate(seeds)
            }
            for f in as_completed(futures):
                yield f.result()

    try:
        yield from _run_executor()
    except RuntimeError as e:
        if "bootstrapping" in str(e) and mp_ctx.get_start_method() == "spawn":
            raise RuntimeError(
                "Pathfinder with mp_ctx='spawn' requires the entry point to be "
                "guarded with `if __name__ == '__main__':`. When using spawn, "
                "wrap your fit_pathfinder call (or the function that contains it) "
                "in that block. See: "
                "https://docs.python.org/3/library/multiprocessing.html"
                "#the-spawn-and-forkserver-start-methods"
            ) from e
        raise
    finally:
        for c in child_conns:
            try:
                c.close()
            except Exception:
                pass
        listener.join(timeout=5)


def _execute_serially(
    fn: SinglePathfinderFn,
    seeds: list[int],
    progress_callbacks: list[Callable | None] | None = None,
) -> Iterator["PathfinderResult"]:
    """Execute pathfinder runs serially."""
    callbacks = progress_callbacks or [None] * len(seeds)
    for seed, cb in zip(seeds, callbacks):
        yield fn(seed, cb)


def make_generator(
    concurrent: Literal["process"] | None,
    fn: SinglePathfinderFn,
    seeds: list[int],
    cores: int | None = None,
    blas_cores: int | None | Literal["auto"] = "auto",
    progress_callbacks: list[Callable | None] | None = None,
    mp_ctx: mp.context.BaseContext | str | None = None,
) -> Iterator["PathfinderResult"]:
    """Generator for executing pathfinder runs concurrently or serially."""
    if concurrent is not None:
        num_paths = len(seeds)
        effective_cores = _default_cores(num_paths, cores)
        _, _, num_blas_per_worker = setup_cores_blas_cores(
            blas_cores,
            num_paths,
            effective_cores,
            _initialize_multiprocessing_context(mp_ctx, quiet=True),
        )
        yield from _execute_concurrently(
            fn,
            seeds,
            cores=effective_cores,
            blas_cores=num_blas_per_worker,
            progress_callbacks=progress_callbacks,
            mp_ctx=mp_ctx,
        )
    else:
        yield from _execute_serially(fn, seeds, progress_callbacks)


@dataclass(slots=True, frozen=True)
class PathfinderResult:
    """
    container for storing results from a single pathfinder run.

    Attributes
    ----------
        samples: posterior samples (1, M, N)
        logP: log probability of model (1, M)
        logQ: log probability of approximation (1, M)
        lbfgs_niter: number of lbfgs iterations (1,)
        elbo_argmax: elbo values at convergence (1,)
        lbfgs_status: LBFGS status
        path_status: path status

    where:
        M: number of samples
        N: number of parameters
    """

    samples: NDArray | None = None
    logP: NDArray | None = None
    logQ: NDArray | None = None
    lbfgs_niter: NDArray | None = None
    elbo_argmax: NDArray | None = None
    inv_hessian_diag: NDArray | None = None
    lbfgs_status: LBFGSStatus = LBFGSStatus.LBFGS_FAILED
    path_status: PathStatus = PathStatus.PATH_FAILED


@dataclass(frozen=True)
class PathfinderConfig:
    """configuration parameters for a single pathfinder"""

    num_draws: int  # same as num_draws_per_path
    maxcor: int
    maxiter: int
    ftol: float
    gtol: float
    maxls: int
    jitter: float
    epsilon: float
    num_elbo_draws: int


@dataclass(slots=True, frozen=True)
class MultiPathfinderResult:
    """
    container for aggregating results from multiple paths.

    Attributes
    ----------
        samples: posterior samples (S, M, N)
        logP: log probability of model (S, M)
        logQ: log probability of approximation (S, M)
        lbfgs_niter: number of lbfgs iterations (S,)
        elbo_argmax: elbo values at convergence (S,)
        lbfgs_status: counter for LBFGS status occurrences
        path_status: counter for path status occurrences
        importance_sampling: importance sampling method used
        warnings: list of warnings
        pareto_k
        pathfinder_config: pathfinder configuration
        compile_time
        compute_time
    where:
        S: number of successful paths, where S <= num_paths
        M: number of samples per path
        N: number of parameters
    """

    samples: NDArray | None = None
    logP: NDArray | None = None
    logQ: NDArray | None = None
    lbfgs_niter: NDArray | None = None
    elbo_argmax: NDArray | None = None
    inv_hessian_diag: NDArray | None = None
    lbfgs_status: Counter = field(default_factory=Counter)
    path_status: Counter = field(default_factory=Counter)
    importance_sampling: str | None = "psis"
    warnings: list[str] = field(default_factory=list)
    pareto_k: float | None = None

    # config
    num_paths: int | None = None
    num_draws: int | None = None
    pathfinder_config: PathfinderConfig | None = None

    # timing
    compile_time: float | None = None
    compute_time: float | None = None

    all_paths_failed: bool = False  # raises ValueError if all paths failed

    @classmethod
    def from_path_results(cls, path_results: list[PathfinderResult]) -> "MultiPathfinderResult":
        """aggregate successful pathfinder results and count the occurrences of each status in PathStatus and LBFGSStatus"""

        NUMERIC_ATTRIBUTES = [
            "samples",
            "logP",
            "logQ",
            "lbfgs_niter",
            "elbo_argmax",
            "inv_hessian_diag",
        ]

        success_results = []
        mpr = cls()

        for pr in path_results:
            if pr.path_status not in FAILED_PATH_STATUS:
                success_results.append(tuple(getattr(pr, attr) for attr in NUMERIC_ATTRIBUTES))

            mpr.lbfgs_status[pr.lbfgs_status] += 1
            mpr.path_status[pr.path_status] += 1

        warnings = _get_status_warning(mpr)

        if success_results:
            results_arr = [np.asarray(x) for x in zip(*success_results)]
            return cls(
                *[np.concatenate(x) if x.ndim > 1 else x for x in results_arr],
                lbfgs_status=mpr.lbfgs_status,
                path_status=mpr.path_status,
                warnings=warnings,
            )
        else:
            return cls(
                lbfgs_status=mpr.lbfgs_status,
                path_status=mpr.path_status,
                warnings=warnings,
                all_paths_failed=True,  # raises ValueError later
            )

    def with_timing(self, compile_time: float, compute_time: float) -> Self:
        """add timing information"""
        return replace(self, compile_time=compile_time, compute_time=compute_time)

    def with_pathfinder_config(self, config: PathfinderConfig) -> Self:
        """add pathfinder configuration"""
        return replace(self, pathfinder_config=config)

    def with_warnings(self, warnings: list[str]) -> Self:
        """add warnings"""
        return replace(self, warnings=warnings)

    def with_importance_sampling(
        self,
        num_draws: int,
        method: Literal["psis", "psir", "identity"] | None,
        random_seed: int | None = None,
    ) -> Self:
        """perform importance sampling"""
        if not self.all_paths_failed:
            isres = _importance_sampling(
                samples=self.samples,
                logP=self.logP,
                logQ=self.logQ,
                num_draws=num_draws,
                method=method,
                random_seed=random_seed,
            )
            return replace(
                self,
                samples=isres.samples,
                importance_sampling=method,
                warnings=[*self.warnings, *isres.warnings],
                pareto_k=isres.pareto_k,
            )
        else:
            return self

    def create_summary(self) -> Table:
        """create rich table summary of pathfinder results"""
        table = Table(
            title="Pathfinder Results",
            title_style="none",
            title_justify="left",
            show_header=False,
            box=None,
            padding=(0, 2),
            show_edge=False,
        )
        table.add_column("Description")
        table.add_column("Value")

        # model info
        if self.samples is not None:
            table.add_row("")
            table.add_row("No. model parameters", str(self.samples.shape[-1]))

        # config
        if self.pathfinder_config is not None:
            table.add_row("")
            table.add_row("Configuration:")
            table.add_row("num_draws_per_path", str(self.pathfinder_config.num_draws))
            table.add_row("history size (maxcor)", str(self.pathfinder_config.maxcor))
            table.add_row("max iterations", str(self.pathfinder_config.maxiter))
            table.add_row("ftol", f"{self.pathfinder_config.ftol:.2e}")
            table.add_row("gtol", f"{self.pathfinder_config.gtol:.2e}")
            table.add_row("max line search", str(self.pathfinder_config.maxls))
            table.add_row("jitter", f"{self.pathfinder_config.jitter}")
            table.add_row("epsilon", f"{self.pathfinder_config.epsilon:.2e}")
            table.add_row("ELBO draws", str(self.pathfinder_config.num_elbo_draws))

        # lbfgs
        table.add_row("")
        table.add_row("LBFGS Status:")
        for status, count in self.lbfgs_status.items():
            table.add_row(str(status.name), str(count))

        if self.lbfgs_niter is not None:
            table.add_row(
                "L-BFGS iterations",
                f"mean {np.mean(self.lbfgs_niter):.0f} ± std {np.std(self.lbfgs_niter):.0f}",
            )

        # paths
        table.add_row("")
        table.add_row("Path Status:")
        for status, count in self.path_status.items():
            table.add_row(str(status.name), str(count))

        if self.elbo_argmax is not None:
            table.add_row(
                "ELBO argmax",
                f"mean {np.mean(self.elbo_argmax):.0f} ± std {np.std(self.elbo_argmax):.0f}",
            )

        # importance sampling section
        if not self.all_paths_failed:
            table.add_row("")
            table.add_row("Importance Sampling:")
            table.add_row("Method", self.importance_sampling)
            if self.pareto_k is not None:
                table.add_row("Pareto k", f"{self.pareto_k:.2f}")

        if self.compile_time is not None:
            table.add_row("")
            table.add_row("Timing (seconds):")
            table.add_row("Compile", f"{self.compile_time:.2f}")

        if self.compute_time is not None:
            table.add_row("Compute", f"{self.compute_time:.2f}")

        if self.compile_time is not None and self.compute_time is not None:
            table.add_row("Total", f"{self.compile_time + self.compute_time:.2f}")

        return table

    def display_summary(self) -> None:
        """display summary including warnings"""
        console = Console()
        summary = self.create_summary()

        # warning messages
        if self.warnings:
            warning_text = [
                Text(),  # blank line
                Text("Warnings:"),
                *(
                    Padding(
                        Text("- " + warning, no_wrap=False).wrap(console, width=console.width - 6),
                        (0, 0, 0, 2),  # left padding only
                    )
                    for warning in self.warnings
                ),
            ]
            output = Group(summary, *warning_text)
        else:
            output = summary

        console.print(output)


def _get_status_warning(mpr: MultiPathfinderResult) -> list[str]:
    """get list of relevant LBFGSStatus and PathStatus warnings given a MultiPathfinderResult"""
    warnings = []

    lbfgs_status_message = {
        LBFGSStatus.MAX_ITER_REACHED: "MAX_ITER_REACHED: LBFGS maximum number of iterations reached. Consider increasing maxiter if this occurence is high relative to the number of paths.",
        LBFGSStatus.INIT_FAILED: "INIT_FAILED: LBFGS failed to initialize. Consider reparameterizing the model or reducing jitter if this occurence is high relative to the number of paths.",
        LBFGSStatus.NON_FINITE: "NON_FINITE: LBFGS objective function produced inf or nan at the last iteration. Consider reparameterizing the model or adjusting the pathfinder arguments if this occurence is high relative to the number of paths.",
        LBFGSStatus.LOW_UPDATE_PCT: "LOW_UPDATE_PCT: Majority of LBFGS iterations were not accepted due to the either: (1) LBFGS function or gradient values containing too many inf or nan values or (2) gradient changes being significantly large, set by epsilon. Consider reparameterizing the model, adjusting initvals or jitter or other pathfinder arguments if this occurence is high relative to the number of paths.",
        LBFGSStatus.INIT_FAILED_LOW_UPDATE_PCT: "INIT_FAILED_LOW_UPDATE_PCT: LBFGS failed to initialize due to the either: (1) LBFGS function or gradient values containing too many inf or nan values or (2) gradient changes being significantly large, set by epsilon. Consider reparameterizing the model, adjusting initvals or jitter or other pathfinder arguments if this occurence is high relative to the number of paths.",
    }

    path_status_message = {
        PathStatus.ELBO_ARGMAX_AT_ZERO: "ELBO_ARGMAX_AT_ZERO: ELBO argmax at zero refers to the first iteration during LBFGS. A high occurrence suggests the model's default initial point + jitter values are concentrated in high-density regions in the target distribution and may result in poor exploration of the parameter space. Consider increasing jitter if this occurrence is high relative to the number of paths.",
        PathStatus.INVALID_LOGQ: "INVALID_LOGQ: Invalid logQ values occur when a path's logQ values are not finite. The failed path is not included in samples when importance sampling is used. Consider reparameterizing the model or adjusting the pathfinder arguments if this occurence is high relative to the number of paths.",
        PathStatus.SINGLE_STEP: "SINGLE_STEP: Pathfinder requires at least two LBFGS steps on a path. A path with only one step produces an invalid result. Consider adjusting initvals/jitter if this occurs.",
    }

    for lbfgs_status in mpr.lbfgs_status:
        if lbfgs_status in lbfgs_status_message:
            warnings.append(lbfgs_status_message.get(lbfgs_status))

    for path_status in mpr.path_status:
        if path_status in path_status_message:
            warnings.append(path_status_message.get(path_status))

    return warnings


def _make_multipath_progress(progressbar: bool) -> CustomProgress:
    return CustomProgress(
        TextColumn("{task.description}", table_column=Column("Path", min_width=7, no_wrap=True)),
        TextColumn(
            "{task.fields[status]}", table_column=Column("Status", min_width=10, no_wrap=True)
        ),
        TextColumn(
            "{task.fields[lbfgs_steps]}", table_column=Column("Steps", min_width=6, no_wrap=True)
        ),
        TextColumn(
            "{task.fields[steps_per_sec]}",
            table_column=Column("Steps/s", min_width=8, no_wrap=True),
        ),
        TextColumn(
            "{task.fields[best_ind]}", table_column=Column("Best step", min_width=9, no_wrap=True)
        ),
        TextColumn(
            "{task.fields[best_elbo]}", table_column=Column("Best ELBO", min_width=12, no_wrap=True)
        ),
        TextColumn(
            "{task.fields[current_elbo]}",
            table_column=Column("Cur ELBO", min_width=12, no_wrap=True),
        ),
        TextColumn(
            "{task.fields[step_size]}",
            table_column=Column("Step size", min_width=10, no_wrap=True),
        ),
        TimeElapsedColumn(table_column=Column("Elapsed", min_width=8, no_wrap=True)),
        include_headers=True,
        console=Console(theme=default_progress_theme),
        disable=not progressbar,
    )


def _make_progress_callback(progress: CustomProgress, task_id: int) -> Callable[[dict], None]:
    def cb(info: dict) -> None:
        fields: dict[str, Any] = {}
        if "status" in info and info["status"] is not None:
            fields["status"] = info["status"]
        if "lbfgs_steps" in info:
            fields["lbfgs_steps"] = info["lbfgs_steps"]
        if "best_elbo" in info:
            val = info["best_elbo"]
            fields["best_elbo"] = (
                f"{val:.3f}" if val is not None and np.isfinite(float(val)) else "—"
            )
        if "best_ind" in info:
            val = info["best_ind"]
            fields["best_ind"] = (
                str(int(val)) if val is not None and np.isfinite(float(val)) else "—"
            )
        if "current_elbo" in info:
            val = info["current_elbo"]
            fields["current_elbo"] = (
                f"{val:.3f}" if val is not None and np.isfinite(float(val)) else "—"
            )
        if "step_size" in info:
            val = info["step_size"]
            fields["step_size"] = (
                f"{val:.2e}" if val is not None and np.isfinite(float(val)) else "—"
            )
        if "steps_per_sec" in info:
            val = info["steps_per_sec"]
            fields["steps_per_sec"] = (
                f"{val:.1f}/s" if val is not None and np.isfinite(float(val)) else "—"
            )
        if fields:
            progress.update(task_id, **fields)
        if info.get("status") in ("ok", "elbo@0"):
            progress.stop_task(task_id)

    return cb


def multipath_pathfinder(
    model: Model,
    num_paths: int,
    num_draws: int,
    num_draws_per_path: int,
    maxcor: int,
    maxiter: int,
    ftol: float,
    gtol: float,
    maxls: int,
    num_elbo_draws: int,
    jitter: float,
    epsilon: float,
    importance_sampling: Literal["psis", "psir", "identity"] | None,
    progressbar: bool,
    concurrent: Literal["process"] | None = "process",
    cores: int | None = None,
    blas_cores: int | None | Literal["auto"] = "auto",
    mp_ctx: mp.context.BaseContext | str | None = None,
    random_seed: RandomSeed = None,
    max_init_retries: int = 10,
    pathfinder_kwargs: dict[str, Any] = {},
    compile_kwargs: dict[str, Any] = {},
    display_summary: bool = True,
) -> MultiPathfinderResult:
    """
    Fit the Pathfinder Variational Inference algorithm using multiple paths with PyMC/PyTensor backend.

    Parameters
    ----------
    model : pymc.Model
        The PyMC model to fit the Pathfinder algorithm to.
    num_paths : int
        Number of independent paths to run in the Pathfinder algorithm. (default is 4) It is recommended to increase num_paths when increasing the jitter value.
    num_draws : int, optional
        Total number of samples to draw from the fitted approximation (default is 1000).
    num_draws_per_path : int, optional
        Number of samples to draw per path (default is 1000).
    maxcor : int, optional
        Maximum number of variable metric corrections used to define the limited memory matrix (default is None). If None, maxcor is set to ceil(3 * log(N)) or 5 whichever is greater, where N is the number of model parameters.
    maxiter : int, optional
        Maximum number of iterations for the L-BFGS optimisation (default is 1000).
    ftol : float, optional
        Tolerance for the decrease in the objective function (default is 1e-5).
    gtol : float, optional
        Tolerance for the norm of the gradient (default is 1e-8).
    maxls : int, optional
        Maximum number of line search steps for the L-BFGS algorithm (default is 1000).
    num_elbo_draws : int, optional
        Number of draws for the Evidence Lower Bound (ELBO) estimation (default is 10).
    jitter : float, optional
        Amount of jitter to apply to initial points (default is 2.0). Note that Pathfinder may be highly sensitive to the jitter value. It is recommended to increase num_paths when increasing the jitter value.
    epsilon: float
        value used to filter out large changes in the direction of the update gradient at each iteration l in L. Iteration l is only accepted if s·z >= epsilon * ||z||² for each l in L. Matches Zhang et al. (2022) Algorithm 3 with default epsilon=1e-12.
    importance_sampling : str, None, optional
        Method to apply sampling based on log importance weights (logP - logQ).
        "psis" : Pareto Smoothed Importance Sampling (default)
                Recommended for more stable results.
        "psir" : Pareto Smoothed Importance Resampling
                Less stable than PSIS.
        "identity" : Applies log importance weights directly without resampling.
        None : No importance sampling weights. Returns raw samples of size (num_paths, num_draws_per_path, N) where N is number of model parameters. Other methods return samples of size (num_draws, N).
    progressbar : bool, optional
        Whether to display a progress bar (default is False). Setting this to True will likely increase the computation time.
    random_seed : RandomSeed, optional
        Random seed for reproducibility.
    postprocessing_backend : str, optional
        Backend for postprocessing transformations, either "cpu" or "gpu" (default is "cpu"). This is only relevant if inference_backend is "blackjax".
    inference_backend : str, optional
        Backend for inference, either "pymc" or "blackjax" (default is "pymc").
    concurrent : str, optional
        How to run paths: ``"process"`` (default) spawns separate worker processes for true
        parallelism, matching PyMC's approach for parallel chains.  Set to ``None`` for serial
        execution (useful for debugging).
    cores : int, optional
        Number of paths to run in parallel. If ``None``, set to min(4, cpu_count, num_paths),
        mirroring pm.sample.
    blas_cores : int or "auto" or None, optional
        Total number of threads BLAS/OpenMP should use per worker. ``"auto"`` (default)
        matches total to ``cores``. ``None`` keeps default BLAS behavior.
    max_init_retries : int, optional
        Maximum number of re-jitter retries per path when LBFGSInitFailed is raised (default is 10).
    pathfinder_kwargs
        Additional keyword arguments for the Pathfinder algorithm. Supported keys:
        ``jacobian`` (bool, default True), ``vectorize`` (bool, default False).
    compile_kwargs
        Additional keyword arguments for the PyTensor compiler. If not provided, a
        performant default is used.

    Returns
    -------
    MultiPathfinderResult
        The result containing samples and other information from the Multi-Path Pathfinder algorithm.
    """

    *path_seeds, choice_seed = _get_seeds_per_chain(random_seed, num_paths + 1)

    pathfinder_config = PathfinderConfig(
        num_draws=num_draws_per_path,
        maxcor=maxcor,
        maxiter=maxiter,
        ftol=ftol,
        gtol=gtol,
        maxls=maxls,
        num_elbo_draws=num_elbo_draws,
        jitter=jitter,
        epsilon=epsilon,
    )

    compile_start = time.time()
    single_pathfinder_fn = make_single_pathfinder_fn(
        model,
        **asdict(pathfinder_config),
        max_init_retries=max_init_retries,
        pathfinder_kwargs=pathfinder_kwargs,
        compile_kwargs=compile_kwargs,
    )
    compile_end = time.time()

    results = []
    compute_start = time.time()
    try:
        # Per-path progress bar (one row per path, updated in real time)
        progress = _make_multipath_progress(progressbar)

        # Create one task per path and build per-path progress callbacks
        task_ids = []
        path_callbacks: list[Callable | None] = []
        with progress:
            for i in range(num_paths):
                tid = progress.add_task(
                    f"Path {i + 1}",
                    status="queued",
                    lbfgs_steps=0,
                    steps_per_sec="—",
                    best_elbo="—",
                    best_ind="—",
                    current_elbo="—",
                    step_size="—",
                    total=None,
                )
                task_ids.append(tid)
                path_callbacks.append(_make_progress_callback(progress, tid))

            # concurrent="process" gives true parallelism via separate worker processes
            # (matching PyMC's approach). concurrent=None is serial (useful for debugging).
            generator = make_generator(
                concurrent=concurrent,
                fn=single_pathfinder_fn,
                seeds=path_seeds,
                cores=cores,
                blas_cores=blas_cores,
                progress_callbacks=path_callbacks,
                mp_ctx=mp_ctx,
            )

            for result in generator:
                try:
                    if isinstance(result, Exception):
                        raise result
                    else:
                        results.append(result)
                except Exception as e:
                    logger.warning("Unexpected error in a path: %s", str(e))
                    results.append(
                        PathfinderResult(
                            path_status=PathStatus.PATH_FAILED,
                            lbfgs_status=LBFGSStatus.LBFGS_FAILED,
                        )
                    )
    except (KeyboardInterrupt, StopIteration) as e:
        # User is free to abort early — MultiPathfinderResult collects all successful results so far.
        if isinstance(e, StopIteration):
            logger.info(str(e))
    finally:
        compute_end = time.time()
        if results:
            mpr = (
                MultiPathfinderResult.from_path_results(results)
                .with_pathfinder_config(config=pathfinder_config)
                .with_importance_sampling(
                    num_draws=num_draws, method=importance_sampling, random_seed=choice_seed
                )
                .with_timing(
                    compile_time=compile_end - compile_start,
                    compute_time=compute_end - compute_start,
                )
            )
            # Display summary conditionally
            if display_summary:
                mpr.display_summary()
            if mpr.all_paths_failed:
                raise ValueError(
                    "All paths failed. Consider decreasing the jitter or reparameterizing the model."
                )
        else:
            raise ValueError(
                "BUG: Failed to iterate!"
                "Please report this issue at: "
                "https://github.com/pymc-devs/pymc-extras/issues "
                "with your code to reproduce the issue and the following details:\n"
                f"pathfinder_config: \n{pathfinder_config}\n"
                f"compile_kwargs: {compile_kwargs}\n"
                f"pathfinder_kwargs: {pathfinder_kwargs}\n"
                f"num_paths: {num_paths}\n"
                f"num_draws: {num_draws}\n"
            )

    return mpr


def fit_pathfinder(
    model=None,
    num_paths: int = 4,  # I
    num_draws: int = 1000,  # R
    num_draws_per_path: int = 1000,  # M
    maxcor: int | None = None,  # J
    maxiter: int = 1000,  # L^max
    ftol: float = 1e-5,
    gtol: float = 1e-8,
    maxls: int = 1000,
    num_elbo_draws: int = 10,  # K
    max_init_retries: int = 10,
    jitter: float = 2.0,
    epsilon: float = 1e-12,
    importance_sampling: Literal["psis", "psir", "identity"] | None = "psis",
    progressbar: bool = True,
    concurrent: Literal["process"] | None = "process",
    cores: int | None = None,
    blas_cores: int | None | Literal["auto"] = "auto",
    mp_ctx: mp.context.BaseContext | str | None = None,
    random_seed: RandomSeed | None = None,
    postprocessing_backend: Literal["cpu", "gpu"] = "cpu",
    inference_backend: Literal["pymc", "blackjax"] = "pymc",
    pathfinder_kwargs: dict[str, Any] = {},
    compile_kwargs: dict[str, Any] = {},
    initvals: dict[str, Any] | None = None,
    # New pathfinder result integration options
    add_pathfinder_groups: bool = True,
    display_summary: bool | Literal["auto"] = "auto",
    store_diagnostics: bool = False,
) -> az.InferenceData:
    """
    Fit the Pathfinder Variational Inference algorithm.

    This function fits the Pathfinder algorithm to a given PyMC model, allowing for multiple paths and draws. It supports both PyMC and BlackJAX backends.

    Parameters
    ----------
    model : pymc.Model
        The PyMC model to fit the Pathfinder algorithm to.
    num_paths : int
        Number of independent paths to run in the Pathfinder algorithm. (default is 4) It is recommended to increase num_paths when increasing the jitter value.
    num_draws : int, optional
        Total number of samples to draw from the fitted approximation (default is 1000).
    num_draws_per_path : int, optional
        Number of samples to draw per path (default is 1000).
    maxcor : int, optional
        Maximum number of variable metric corrections used to define the limited memory matrix (default is None). If None, maxcor is set to ceil(3 * log(N)) or 5 whichever is greater, where N is the number of model parameters.
    maxiter : int, optional
        Maximum number of iterations for the L-BFGS optimisation (default is 1000).
    ftol : float, optional
        Tolerance for the decrease in the objective function (default is 1e-5).
    gtol : float, optional
        Tolerance for the norm of the gradient (default is 1e-8).
    maxls : int, optional
        Maximum number of line search steps for the L-BFGS algorithm (default is 1000).
    num_elbo_draws : int, optional
        Number of draws for the Evidence Lower Bound (ELBO) estimation (default is 10).
    jitter : float, optional
        Amount of jitter to apply to initial points (default is 2.0). Note that Pathfinder may be highly sensitive to the jitter value. It is recommended to increase num_paths when increasing the jitter value.
    epsilon: float
        value used to filter out large changes in the direction of the update gradient at each iteration l in L. Iteration l is only accepted if s·z >= epsilon * ||z||² for each l in L. Matches Zhang et al. (2022) Algorithm 3 with default epsilon=1e-12.
    importance_sampling : str, None, optional
        Method to apply sampling based on log importance weights (logP - logQ).
        Options are:

        - "psis" : Pareto Smoothed Importance Sampling (default). Usually most stable.
        - "psir" : Pareto Smoothed Importance Resampling. Less stable than PSIS.
        - "identity" : Applies log importance weights directly without resampling.
        - None : No importance sampling weights. Returns raw samples of size (num_paths, num_draws_per_path, N) where N is number of model parameters. Other methods return samples of size (num_draws, N).

    progressbar : bool, optional
        Whether to display a progress bar (default is True). Setting this to False will likely reduce the computation time.
    random_seed : RandomSeed, optional
        Random seed for reproducibility.
    postprocessing_backend : str, optional
        Backend for postprocessing transformations, either "cpu" or "gpu" (default is "cpu"). This is only relevant if inference_backend is "blackjax".
    inference_backend : str, optional
        Backend for inference, either "pymc" or "blackjax" (default is "pymc").
    concurrent : str, optional
        How to run paths: ``"process"`` (default) spawns separate worker processes for true
        parallelism, matching PyMC's approach for parallel chains.  Set to ``None`` for serial
        execution (useful for debugging).
    cores : int, optional
        Number of paths to run in parallel. If ``None``, set to min(4, cpu_count, num_paths),
        mirroring pm.sample.
    blas_cores : int or "auto" or None, optional
        Total number of threads BLAS/OpenMP should use per worker. ``"auto"`` (default)
        matches total to ``cores``. ``None`` keeps default BLAS behavior.
    mp_ctx : str or multiprocessing.Context, optional
        Multiprocessing context for parallel path execution (e.g. ``"spawn"``, ``"fork"``).
    max_init_retries : int, optional
        Maximum number of re-jitter retries per path when the initial point fails (default is 10).
    pathfinder_kwargs : dict, optional
        Additional keyword arguments for the Pathfinder algorithm. Supported keys:
        - ``vectorize`` (bool, default False): If True, use ``vectorize_graph``
          instead of ``pytensor.map`` for batched log-probability evaluation.
          May be faster for small models but may use more memory for large
          parameter counts.
        - ``jacobian`` (bool, default True): Whether to include the Jacobian
          determinant in the log-probability computation. Setting to False is not
          recommended and may result in very high Pareto k values.

    compile_kwargs
        Additional keyword arguments for the PyTensor compiler. If not provided, a
        performant default is used.
    initvals: dict | None = None
        Initial values for the model parameters, as str:ndarray key-value pairs. Paritial initialization is permitted.
        If None, the model's default initial values are used.
    add_pathfinder_groups : bool, optional
        Whether to add pathfinder results as additional groups to the InferenceData (default is True).
        When True, adds pathfinder and pathfinder_paths groups with optimization diagnostics.
    display_summary : bool or "auto", optional
        Whether to display the pathfinder results summary (default is "auto").
        "auto" preserves current behavior, False suppresses console output.
    store_diagnostics : bool, optional
        Whether to include potentially large diagnostic arrays in the pathfinder groups (default is False).
    pathfinder_group : str, optional
        Name for the main pathfinder results group (default is "sample_stats").

    Returns
    -------
    :class:`~arviz.InferenceData`
        The inference data containing the results of the Pathfinder algorithm.

    References
    ----------
    Zhang, L., Carpenter, B., Gelman, A., & Vehtari, A. (2022). Pathfinder: Parallel quasi-Newton variational inference. Journal of Machine Learning Research, 23(306), 1-49.
    """

    model = modelcontext(model)
    if concurrent is not None and concurrent != "process":
        raise ValueError(
            f"concurrent must be 'process' or None, got {concurrent!r}. "
            "Thread-based parallelism is not supported: PyTensor compiled functions "
            "share internal storage and would corrupt each other's state."
        )

    if initvals is not None:
        model = pm.model.fgraph.clone_model(model)  # Create a clone of the model
        for (
            rv_name,
            ivals,
        ) in initvals.items():  # Set the initial values for the variables in the clone
            model.set_initval(model.named_vars[rv_name], ivals)

    valid_importance_sampling = {"psis", "psir", "identity", None}

    if importance_sampling is not None:
        importance_sampling = importance_sampling.lower()

    if importance_sampling not in valid_importance_sampling:
        raise ValueError(f"Invalid importance sampling method: {importance_sampling}")

    N = DictToArrayBijection.map(model.initial_point()).data.shape[0]

    if maxcor is None:
        # Based on tests, this seems to be a good default value. Higher maxcor values do not necessarily lead to better results and can slow down the algorithm. Also, if results do benefit from a higher maxcor value, the improvement may be diminishing w.r.t. the increase in maxcor.
        maxcor = np.ceil(3 * np.log(N)).astype(np.int32)
        maxcor = max(maxcor, 5)

    # Handle display_summary logic
    should_display_summary = display_summary == "auto" or display_summary is True

    if inference_backend == "pymc":
        mp_result = multipath_pathfinder(
            model,
            num_paths=num_paths,
            num_draws=num_draws,
            num_draws_per_path=num_draws_per_path,
            maxcor=maxcor,
            maxiter=maxiter,
            ftol=ftol,
            gtol=gtol,
            maxls=maxls,
            num_elbo_draws=num_elbo_draws,
            jitter=jitter,
            epsilon=epsilon,
            importance_sampling=importance_sampling,
            progressbar=progressbar,
            concurrent=concurrent,
            cores=cores,
            blas_cores=blas_cores,
            mp_ctx=mp_ctx,
            max_init_retries=max_init_retries,
            random_seed=random_seed,
            pathfinder_kwargs=pathfinder_kwargs,
            compile_kwargs=compile_kwargs,
            display_summary=should_display_summary,
        )
        pathfinder_samples = mp_result.samples
    elif inference_backend == "blackjax":
        import blackjax
        import jax

        if version.parse(blackjax.__version__).major < 1:
            raise ImportError("fit_pathfinder requires blackjax 1.0 or above")

        _, pathfinder_seed, sample_seed = _get_seeds_per_chain(random_seed, 3)
        # TODO: extend initial points with jitter_scale to blackjax
        # TODO: extend blackjax pathfinder to multiple paths
        x0, _ = DictToArrayBijection.map(model.initial_point())
        logp_func = get_jaxified_logp_of_ravel_inputs(model)
        pathfinder_state, _ = blackjax.vi.pathfinder.approximate(
            rng_key=jax.random.key(pathfinder_seed),
            logdensity_fn=logp_func,
            initial_position=x0,
            num_samples=num_elbo_draws,
            maxiter=maxiter,
            maxcor=maxcor,
            maxls=maxls,
            ftol=ftol,
            gtol=gtol,
            **pathfinder_kwargs,
        )
        pathfinder_samples, _ = blackjax.vi.pathfinder.sample(
            rng_key=jax.random.key(sample_seed),
            state=pathfinder_state,
            num_samples=num_draws,
        )
    else:
        raise ValueError(f"Invalid inference_backend: {inference_backend}")

    logger.info("Transforming variables...")

    idata = convert_flat_trace_to_idata(
        pathfinder_samples,
        postprocessing_backend=postprocessing_backend,
        inference_backend=inference_backend,
        model=model,
        importance_sampling=importance_sampling,
    )

    idata = add_data_to_inference_data(idata, progressbar, model, compile_kwargs)

    # Add pathfinder results to InferenceData if requested
    if add_pathfinder_groups:
        if inference_backend == "pymc":
            from pymc_extras.inference.pathfinder.idata import add_pathfinder_to_inference_data

            idata = add_pathfinder_to_inference_data(
                idata=idata,
                result=mp_result,
                model=model,
                store_diagnostics=store_diagnostics,
            )
        else:
            warnings.warn(
                f"Pathfinder diagnostic groups are only supported with the PyMC backend. "
                f"Current backend is '{inference_backend}', which does not support adding "
                "pathfinder diagnostics to InferenceData. The InferenceData will only contain "
                "posterior samples. To add diagnostic groups, use inference_backend='pymc', "
                "or set add_pathfinder_groups=False to suppress this warning.",
                UserWarning,
                stacklevel=2,
            )

    return idata



import itertools

from contextlib import suppress as does_not_warn

import numpy as np
import pandas as pd
import pymc as pm
import pytensor.tensor as pt
import pytest

from arviz import InferenceData, dict_to_dataset
from pymc import Model, draw
from pymc.distributions import transforms
from pymc.distributions.transforms import ordered
from pymc.initial_point import make_initial_point_expression
from pymc.pytensorf import constant_fold, inputvars
from pymc.util import UNSET
from scipy.special import log_softmax, logsumexp
from scipy.stats import halfnorm, norm

from pymc_extras.model.marginal.distributions import MarginalRV
from pymc_extras.model.marginal.marginal_model import (
    marginalize,
    recover_marginals,
    unmarginalize,
)
from pymc_extras.utils.model_equivalence import equal_computations_up_to_root, equivalent_models

# FIXME: A Blockwise of Reshape should be rewritten into a Reshape, as it's rather inneficient
# This shows up in `test_one_to_many_unaligned_marginalized_rvs`
pytestmark = pytest.mark.filterwarnings(
    "ignore:Numba will use object mode to run Blockwise{Reshapep*:UserWarning"
)


def test_basic_marginalized_rv():
    data = [2] * 5

    with Model() as m:
        sigma = pm.HalfNormal("sigma")
        idx = pm.Categorical("idx", p=[0.1, 0.3, 0.6])
        mu = pt.switch(
            pt.eq(idx, 0),
            -1.0,
            pt.switch(
                pt.eq(idx, 1),
                0.0,
                1.0,
            ),
        )
        y = pm.Normal("y", mu=mu, sigma=sigma)
        z = pm.Normal("z", y, observed=data)

    marginal_m = marginalize(m, [idx])
    assert isinstance(marginal_m["y"].owner.op, MarginalRV)
    assert ["idx"] not in [rv.name for rv in marginal_m.free_RVs]

    # Test forward draws
    y_draws, z_draws = draw(
        [marginal_m["y"], marginal_m["z"]],
        # Make sigma very small to make draws deterministic
        givens={marginal_m["sigma"]: 0.001},
        draws=1000,
        random_seed=54,
    )
    assert sorted(np.unique(y_draws.round()) == [-1.0, 0.0, 1.0])
    assert z_draws[y_draws < 0].mean() < z_draws[y_draws > 0].mean()

    # Test initial_point
    ips = make_initial_point_expression(
        # Use basic_RVs to include the observed RV
        free_rvs=marginal_m.basic_RVs,
        rvs_to_transforms=marginal_m.rvs_to_transforms,
        initval_strategies={},
    )
    # After simplification, we should have only constants in the graph (expect alloc which isn't constant folded):
    ip_sigma, ip_y, ip_z = constant_fold(ips)
    np.testing.assert_allclose(ip_sigma, 1.0)
    np.testing.assert_allclose(ip_y, 1.0)
    np.testing.assert_allclose(ip_z, np.full((5,), 1.0))

    marginal_ip = marginal_m.initial_point()
    expected_ip = m.initial_point()
    expected_ip.pop("idx")
    assert marginal_ip == expected_ip

    # Test logp
    with pm.Model() as ref_m:
        sigma = pm.HalfNormal("sigma")
        y = pm.NormalMixture("y", w=[0.1, 0.3, 0.6], mu=[-1, 0, 1], sigma=sigma)
        z = pm.Normal("z", y, observed=data)

    np.testing.assert_almost_equal(
        marginal_m.compile_logp()(marginal_ip),
        ref_m.compile_logp()(marginal_ip),
    )
    np.testing.assert_almost_equal(
        marginal_m.compile_dlogp([marginal_m["y"]])(marginal_ip),
        ref_m.compile_dlogp([ref_m["y"]])(marginal_ip),
    )


def test_one_to_one_marginalized_rvs():
    """Test case with multiple, independent marginalized RVs."""
    with Model() as m:
        sigma = pm.HalfNormal("sigma")
        idx1 = pm.Bernoulli("idx1", p=0.75)
        x = pm.Normal("x", mu=idx1, sigma=sigma)
        idx2 = pm.Bernoulli("idx2", p=0.75, shape=(5,))
        y = pm.Normal("y", mu=(idx2 * 2 - 1), sigma=sigma, shape=(5,))

    marginal_m = marginalize(m, [idx1, idx2])
    assert isinstance(marginal_m["x"].owner.op, MarginalRV)
    assert isinstance(marginal_m["y"].owner.op, MarginalRV)
    assert marginal_m["x"].owner is not marginal_m["y"].owner

    with pm.Model() as ref_m:
        sigma = pm.HalfNormal("sigma")
        x = pm.NormalMixture("x", w=[0.25, 0.75], mu=[0, 1], sigma=sigma)
        y = pm.NormalMixture("y", w=[0.25, 0.75], mu=[-1, 1], sigma=sigma, shape=(5,))

    # Test logp
    test_point = ref_m.initial_point()
    x_logp, y_logp = marginal_m.compile_logp(vars=[marginal_m["x"], marginal_m["y"]], sum=False)(
        test_point
    )
    x_ref_log, y_ref_logp = ref_m.compile_logp(vars=[ref_m["x"], ref_m["y"]], sum=False)(test_point)
    np.testing.assert_array_almost_equal(x_logp, x_ref_log.sum())
    np.testing.assert_array_almost_equal(y_logp, y_ref_logp)


def test_one_to_many_marginalized_rvs():
    """Test that marginalization works when there is more than one dependent RV"""
    with Model() as m:
        sigma = pm.HalfNormal("sigma")
        idx = pm.Bernoulli("idx", p=0.75)
        x = pm.Normal("x", mu=idx, sigma=sigma)
        y = pm.Normal("y", mu=(idx * 2 - 1), sigma=sigma, shape=(5,))

    marginal_m = marginalize(m, [idx])

    marginal_x = marginal_m["x"]
    marginal_y = marginal_m["y"]
    assert isinstance(marginal_x.owner.op, MarginalRV)
    assert isinstance(marginal_y.owner.op, MarginalRV)
    assert marginal_x.owner is marginal_y.owner

    ref_logp_x_y_fn = m.compile_logp([idx, x, y])
    tp = marginal_m.initial_point()
    ref_logp_x_y = logsumexp([ref_logp_x_y_fn({**tp, **{"idx": idx}}) for idx in (0, 1)])
    with pytest.warns(UserWarning, match="There are multiple dependent variables"):
        logp_x_y = marginal_m.compile_logp([marginal_x, marginal_y])(tp)
    np.testing.assert_array_almost_equal(logp_x_y, ref_logp_x_y)


def test_one_to_many_unaligned_marginalized_rvs():
    """Test that marginalization works when there is more than one dependent RV with batch dimensions that are not aligned"""

    def build_model(build_batched: bool):
        with Model() as m:
            if build_batched:
                idx = pm.Bernoulli("idx", p=[0.75, 0.4], shape=(3, 2))
            else:
                idxs = [pm.Bernoulli(f"idx_{i}", p=(0.75 if i % 2 == 0 else 0.4)) for i in range(6)]
                idx = pt.stack(idxs, axis=0).reshape((3, 2))

            x = pm.Normal("x", mu=idx.T[:, :, None], shape=(2, 3, 1))
            y = pm.Normal("y", mu=(idx * 2 - 1), shape=(1, 3, 2))

        return m

    marginal_m = marginalize(build_model(build_batched=True), ["idx"])
    ref_m = marginalize(build_model(build_batched=False), [f"idx_{i}" for i in range(6)])

    test_point = marginal_m.initial_point()

    with pytest.warns(UserWarning, match="There are multiple dependent variables"):
        np.testing.assert_allclose(
            marginal_m.compile_logp()(test_point),
            ref_m.compile_logp()(test_point),
        )


def test_many_to_one_marginalized_rvs():
    """Test when random variables depend on multiple marginalized variables"""
    with Model() as m:
        x = pm.Bernoulli("x", 0.1)
        y = pm.Bernoulli("y", 0.3)
        z = pm.DiracDelta("z", c=x + y)

    logp_fn = marginalize(m, [x, y]).compile_logp()

    np.testing.assert_allclose(np.exp(logp_fn({"z": 0})), 0.9 * 0.7)
    np.testing.assert_allclose(np.exp(logp_fn({"z": 1})), 0.9 * 0.3 + 0.1 * 0.7)
    np.testing.assert_allclose(np.exp(logp_fn({"z": 2})), 0.1 * 0.3)


@pytest.mark.parametrize("batched", (False, "left", "right"))
def test_nested_marginalized_rvs(batched):
    """Test that marginalization works when there are nested marginalized RVs"""

    def build_model(build_batched: bool) -> Model:
        idx_shape = (3,) if build_batched else ()
        sub_idx_shape = (5,) if not build_batched else (5, 3) if batched == "left" else (3, 5)

        with Model() as m:
            sigma = pm.HalfNormal("sigma")

            idx = pm.Bernoulli("idx", p=0.75, shape=idx_shape)
            dep = pm.Normal("dep", mu=pt.switch(pt.eq(idx, 0), -1000.0, 1000.0), sigma=sigma)

            sub_idx_p = pt.switch(pt.eq(idx, 0), 0.15, 0.95)
            if build_batched and batched == "right":
                sub_idx_p = sub_idx_p[..., None]
                dep = dep[..., None]
            sub_idx = pm.Bernoulli("sub_idx", p=sub_idx_p, shape=sub_idx_shape)
            sub_dep = pm.Normal("sub_dep", mu=dep + sub_idx * 100, sigma=sigma)

        return m

    marginal_m = marginalize(build_model(build_batched=batched), ["idx", "sub_idx"])
    assert all(rv.name not in ("idx", "sub_idx") for rv in marginal_m.free_RVs)

    # Test forward draws and initial_point, shouldn't depend on batching, so we only test one case
    if not batched:
        # Test forward draws
        dep_draws, sub_dep_draws = draw(
            [marginal_m["dep"], marginal_m["sub_dep"]],
            # Make sigma very small to make draws deterministic
            givens={marginal_m["sigma"]: 0.001},
            draws=1000,
            random_seed=214,
        )
        assert sorted(np.unique(dep_draws.round()) == [-1000.0, 1000.0])
        assert sorted(np.unique(sub_dep_draws.round()) == [-1000.0, -900.0, 1000.0, 1100.0])

        # Test initial_point
        ips = make_initial_point_expression(
            free_rvs=[marginal_m["sigma"], marginal_m["dep"], marginal_m["sub_dep"]],
            rvs_to_transforms=marginal_m.rvs_to_transforms,
            initval_strategies={},
        )
        # After simplification, we should have only constants in the graph
        ip_sigma, ip_dep, ip_sub_dep = constant_fold(ips)
        np.testing.assert_allclose(ip_sigma, 1.0)
        np.testing.assert_allclose(ip_dep, 1000.0)
        np.testing.assert_allclose(ip_sub_dep, np.full((5,), 1100.0))

    # Test logp
    ref_m = build_model(build_batched=False)
    ref_logp_fn = ref_m.compile_logp(
        vars=[ref_m["idx"], ref_m["dep"], ref_m["sub_idx"], ref_m["sub_dep"]]
    )

    test_point = ref_m.initial_point()
    test_point["dep"] = np.full_like(test_point["dep"], 1000)
    test_point["sub_dep"] = np.full_like(test_point["sub_dep"], 1000 + 100)
    ref_logp = logsumexp(
        [
            ref_logp_fn({**test_point, **{"idx": idx, "sub_idx": np.array(sub_idxs)}})
            for idx in (0, 1)
            for sub_idxs in itertools.product((0, 1), repeat=5)
        ]
    )
    if batched:
        ref_logp *= 3

    test_point = marginal_m.initial_point()
    test_point["dep"] = np.full_like(test_point["dep"], 1000)
    test_point["sub_dep"] = np.full_like(test_point["sub_dep"], 1000 + 100)

    with pytest.warns(UserWarning, match="There are multiple dependent variables"):
        logp = marginal_m.compile_logp(vars=[marginal_m["dep"], marginal_m["sub_dep"]])(test_point)

    np.testing.assert_almost_equal(logp, ref_logp)


def test_interdependent_rvs():
    """Test Marginalization when dependent RVs are interdependent."""
    with Model() as m:
        idx = pm.Bernoulli("idx", p=0.75)
        x = pm.Normal("x", mu=idx * 2, sigma=1e-3)
        # Y depends on both x and idx
        y = pm.Normal("y", mu=x * idx * 2, sigma=1e-3)

    marginal_m = marginalize(m, "idx")

    marginal_x = marginal_m["x"]
    marginal_y = marginal_m["y"]
    assert isinstance(marginal_x.owner.op, MarginalRV)
    assert isinstance(marginal_y.owner.op, MarginalRV)
    assert marginal_x.owner is marginal_y.owner

    # Test forward draws
    x_draws, y_draws = draw([marginal_x, marginal_y], draws=1000, random_seed=54)
    assert sorted(np.unique(x_draws.round())) == [0, 2]
    assert sorted(np.unique(y_draws.round())) == [0, 4]
    assert np.unique(y_draws[x_draws < 1].round()) == [0]
    assert np.unique(y_draws[x_draws > 1].round()) == [4]

    # Test initial_point
    ips = make_initial_point_expression(
        free_rvs=[marginal_m["x"], marginal_m["y"]],
        rvs_to_transforms={},
        initval_strategies={},
    )
    # After simplification, we should have only constants in the graph
    ip_x, ip_y = constant_fold(ips)
    np.testing.assert_allclose(ip_x, 2.0)
    np.testing.assert_allclose(ip_y, 4.0)

    # Test custom initval strategy
    ips = make_initial_point_expression(
        # Test that order does not matter
        free_rvs=[marginal_m["y"], marginal_m["x"]],
        rvs_to_transforms={},
        initval_strategies={marginal_x: pt.constant(5.0)},
    )
    ip_y, ip_x = constant_fold(ips)
    np.testing.assert_allclose(ip_x, 5.0)
    np.testing.assert_allclose(ip_y, 10.0)

    # Test logp
    test_point = marginal_m.initial_point()
    ref_logp_fn = m.compile_logp([m["idx"], m["x"], m["y"]])
    ref_logp = logsumexp([ref_logp_fn({**test_point, **{"idx": idx}}) for idx in (0, 1)])
    with pytest.warns(UserWarning, match="There are multiple dependent variables"):
        logp = marginal_m.compile_logp([marginal_m["x"], marginal_m["y"]])(test_point)
    np.testing.assert_almost_equal(logp, ref_logp)


@pytest.mark.parametrize("advanced_indexing", (False, True))
def test_marginalized_index_as_key(advanced_indexing):
    """Test we can marginalize graphs where indexing is used as a mapping."""

    w = [0.1, 0.3, 0.6]
    mu = pt.as_tensor([-1, 0, 1])

    if advanced_indexing:
        y_val = pt.as_tensor([[-1, -1], [0, 1]])
        shape = (2, 2)
    else:
        y_val = -1
        shape = ()

    with Model() as m:
        x = pm.Categorical("x", p=w, shape=shape)
        y = pm.Normal("y", mu[x].T, sigma=1, observed=y_val)

    marginal_m = marginalize(m, x)

    marginal_logp = marginal_m.compile_logp(sum=False)({})[0]
    ref_logp = pm.logp(pm.NormalMixture.dist(w=w, mu=mu.T, sigma=1, shape=shape), y_val).eval()

    np.testing.assert_allclose(marginal_logp, ref_logp)


def test_marginalized_index_as_value_and_key():
    """Test we can marginalize graphs were marginalized_rv is indexed."""

    def build_model(build_batched: bool) -> Model:
        with Model() as m:
            if build_batched:
                latent_state = pm.Bernoulli("latent_state", p=0.3, size=(4,))
            else:
                latent_state = pm.math.stack(
                    [pm.Bernoulli(f"latent_state_{i}", p=0.3) for i in range(4)]
                )
            # latent state is used as the indexed variable
            latent_intensities = pt.where(latent_state[:, None], [0.0, 1.0, 2.0], [0.0, 10.0, 20.0])
            picked_intensity = pm.Categorical("picked_intensity", p=[0.2, 0.2, 0.6])
            # picked intensity is used as the indexing variable
            pm.Normal(
                "intensity",
                mu=latent_intensities[:, picked_intensity],
                observed=[0.5, 1.5, 5.0, 15.0],
            )
        return m

    # We compare with the equivalent but less efficient batched model
    m = build_model(build_batched=True)
    ref_m = build_model(build_batched=False)

    m = marginalize(m, ["latent_state"])
    ref_m = marginalize(ref_m, [f"latent_state_{i}" for i in range(4)])
    test_point = {"picked_intensity": 1}
    np.testing.assert_allclose(
        m.compile_logp()(test_point),
        ref_m.compile_logp()(test_point),
    )

    m = marginalize(m, ["picked_intensity"])
    ref_m = marginalize(ref_m, ["picked_intensity"])
    test_point = {}
    np.testing.assert_allclose(
        m.compile_logp()(test_point),
        ref_m.compile_logp()(test_point),
    )


class TestNotSupportedMixedDims:
    """Test lack of support for models where batch dims of marginalized variables are mixed."""

    def test_mixed_dims_via_transposed_dot(self):
        with Model() as m:
            idx = pm.Bernoulli("idx", p=0.7, shape=2)
            y = pm.Normal("y", mu=idx @ idx.T)

        with pytest.raises(NotImplementedError):
            marginalize(m, idx)

    def test_mixed_dims_via_indexing(self):
        mean = pt.as_tensor([[0.1, 0.9], [0.6, 0.4]])

        with Model() as m:
            idx = pm.Bernoulli("idx", p=0.7, shape=2)
            y = pm.Normal("y", mu=mean[idx, :] + mean[:, idx])
        with pytest.raises(NotImplementedError):
            marginalize(m, idx)

        with Model() as m:
            idx = pm.Bernoulli("idx", p=0.7, shape=2)
            y = pm.Normal("y", mu=mean[idx, None] + mean[None, idx])
        with pytest.raises(NotImplementedError):
            marginalize(m, idx)

        with Model() as m:
            idx = pm.Bernoulli("idx", p=0.7, shape=2)
            mu = pt.specify_broadcastable(mean[:, None][idx], 1) + pt.specify_broadcastable(
                mean[None, :][:, idx], 0
            )
            y = pm.Normal("y", mu=mu)
        with pytest.raises(NotImplementedError):
            marginalize(m, idx)

        with Model() as m:
            idx = pm.Bernoulli("idx", p=0.7, shape=2)
            y = pm.Normal("y", mu=idx[0] + idx[1])
        with pytest.raises(NotImplementedError):
            marginalize(m, idx)

    def test_mixed_dims_via_vector_indexing(self):
        with Model() as m:
            idx = pm.Bernoulli("idx", p=0.7, shape=2)
            y = pm.Normal("y", mu=idx[[0, 1, 0, 0]])
        with pytest.raises(NotImplementedError):
            marginalize(m, idx)

        with Model() as m:
            idx = pm.Categorical("key", p=[0.1, 0.3, 0.6], shape=(2, 2))
            y = pm.Normal("y", pt.as_tensor([[0, 1], [2, 3]])[idx.astype(bool)])
        with pytest.raises(NotImplementedError):
            marginalize(m, idx)

    def test_mixed_dims_via_support_dimension(self):
        with Model() as m:
            x = pm.Bernoulli("x", p=0.7, shape=3)
            y = pm.Dirichlet("y", a=x * 10 + 1)
        with pytest.raises(NotImplementedError):
            marginalize(m, x)

    def test_mixed_dims_via_nested_marginalization(self):
        with Model() as m:
            x = pm.Bernoulli("x", p=0.7, shape=(3,))
            y = pm.Bernoulli("y", p=0.7, shape=(2,))
            z = pm.Normal("z", mu=pt.add.outer(x, y), shape=(3, 2))

        with pytest.raises(NotImplementedError):
            marginalize(m, [x, y])


def test_marginalized_deterministic_and_potential():
    rng = np.random.default_rng(299)

    with Model() as m:
        x = pm.Bernoulli("x", p=0.7)
        y = pm.Normal("y", x)
        z = pm.Normal("z", x)
        det = pm.Deterministic("det", y + z)
        pot = pm.Potential("pot", y + z + 1)

    marginal_m = marginalize(m, [x])

    y_draw, z_draw, det_draw, pot_draw = pm.draw([y, z, det, pot], draws=5, random_seed=rng)
    np.testing.assert_almost_equal(y_draw + z_draw, det_draw)
    np.testing.assert_almost_equal(det_draw, pot_draw - 1)

    y_value = marginal_m.rvs_to_values[marginal_m["y"]]
    z_value = marginal_m.rvs_to_values[marginal_m["z"]]
    det_value, pot_value = marginal_m.replace_rvs_by_values([marginal_m["det"], marginal_m["pot"]])
    assert set(inputvars([det_value, pot_value])) == {y_value, z_value}
    assert det_value.eval({y_value: 2, z_value: 5}) == 7
    assert pot_value.eval({y_value: 2, z_value: 5}) == 8


def test_not_supported_marginalized_deterministic_and_potential():
    with Model() as m:
        x = pm.Bernoulli("x", p=0.7)
        y = pm.Normal("y", x)
        det = pm.Deterministic("det", x + y)

    with pytest.raises(
        NotImplementedError, match="Cannot marginalize x due to dependent Deterministic det"
    ):
        marginalize(m, [x])

    with Model() as m:
        x = pm.Bernoulli("x", p=0.7)
        y = pm.Normal("y", x)
        pot = pm.Potential("pot", x + y)

    with pytest.raises(
        NotImplementedError, match="Cannot marginalize x due to dependent Potential pot"
    ):
        marginalize(m, [x])


@pytest.mark.parametrize(
    "transform, expected_warning",
    (
        (None, does_not_warn()),
        (UNSET, does_not_warn()),
        (transforms.log, does_not_warn()),
        (transforms.Chain([transforms.logodds, transforms.log]), does_not_warn()),
        (
            transforms.Interval(0, 2),
            pytest.warns(
                UserWarning, match="which depends on the marginalized idx may no longer work"
            ),
        ),
        (
            transforms.Chain([transforms.log, transforms.Interval(-1, 1)]),
            pytest.warns(
                UserWarning, match="which depends on the marginalized idx may no longer work"
            ),
        ),
    ),
)
def test_marginalized_transforms(transform, expected_warning):
    w = [0.1, 0.3, 0.6]
    data = [0, 5, 10]
    initval = 0.7  # Value that will be negative on the unconstrained space

    with pm.Model() as m_ref:
        sigma = pm.Mixture(
            "sigma",
            w=w,
            comp_dists=pm.HalfNormal.dist([1, 2, 3]),
            initval=initval,
            default_transform=transform,
        )
        y = pm.Normal("y", 0, sigma, observed=data)

    with Model() as m:
        idx = pm.Categorical("idx", p=w)
        sigma = pm.HalfNormal(
            "sigma",
            pt.switch(
                pt.eq(idx, 0),
                1,
                pt.switch(
                    pt.eq(idx, 1),
                    2,
                    3,
                ),
            ),
            default_transform=transform,
        )
        y = pm.Normal("y", 0, sigma, observed=data)

    with expected_warning:
        marginal_m = marginalize(m, [idx])

    marginal_m.set_initval(marginal_m["sigma"], initval)
    ip = marginal_m.initial_point()
    if transform is not None:
        if transform is UNSET:
            transform_name = "log"
        else:
            transform_name = transform.name
        assert -np.inf < ip[f"sigma_{transform_name}__"] < 0.0
    np.testing.assert_allclose(marginal_m.compile_logp()(ip), m_ref.compile_logp()(ip))


def test_data_container():
    """Test that MarginalModel can handle Data containers."""
    with Model(coords={"obs": [0]}) as m:
        x = pm.Data("x", 2.5)
        idx = pm.Bernoulli("idx", p=0.7, dims="obs")
        y = pm.Normal("y", idx * x, dims="obs")

    marginal_m = marginalize(m, [idx])

    logp_fn = marginal_m.compile_logp()

    with pm.Model(coords={"obs": [0]}) as m_ref:
        x = pm.Data("x", 2.5)
        y = pm.NormalMixture("y", w=[0.3, 0.7], mu=[0, x], dims="obs")

    ref_logp_fn = m_ref.compile_logp()

    for i, x_val in enumerate((-1.5, 2.5, 3.5), start=1):
        for m in (marginal_m, m_ref):
            m.set_dim("obs", new_length=i, coord_values=tuple(range(i)))
            pm.set_data({"x": x_val}, model=m)

        ip = marginal_m.initial_point()
        np.testing.assert_allclose(logp_fn(ip), ref_logp_fn(ip))


def test_mutable_indexing_jax_backend():
    pytest.importorskip("jax")
    from pymc.sampling.jax import get_jaxified_logp

    with Model() as model:
        data = pm.Data("data", np.zeros(10))

        cat_effect = pm.Normal("cat_effect", sigma=1, shape=5)
        cat_effect_idx = pm.Data("cat_effect_idx", np.array([0, 1] * 5))

        is_outlier = pm.Bernoulli("is_outlier", 0.4, shape=10)
        pm.LogNormal("y", mu=cat_effect[cat_effect_idx], sigma=1 + is_outlier, observed=data)
    marginal_model = marginalize(model, ["is_outlier"])
    get_jaxified_logp(marginal_model)


class TestFullModels:
    @pytest.fixture
    def disaster_model(self):
        # fmt: off
        disaster_data = pd.Series(
            [4, 5, 4, 0, 1, 4, 3, 4, 0, 6, 3, 3, 4, 0, 2, 6,
             3, 3, 5, 4, 5, 3, 1, 4, 4, 1, 5, 5, 3, 4, 2, 5,
             2, 2, 3, 4, 2, 1, 3, np.nan, 2, 1, 1, 1, 1, 3, 0, 0,
             1, 0, 1, 1, 0, 0, 3, 1, 0, 3, 2, 2, 0, 1, 1, 1,
             0, 1, 0, 1, 0, 0, 0, 2, 1, 0, 0, 0, 1, 1, 0, 2,
             3, 3, 1, np.nan, 2, 1, 1, 1, 1, 2, 4, 2, 0, 0, 1, 4,
             0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1]
        )
        # fmt: on
        years = np.arange(1851, 1962)

        with Model() as disaster_model:
            switchpoint = pm.DiscreteUniform("switchpoint", lower=years.min(), upper=years.max())
            early_rate = pm.Exponential("early_rate", 1.0)
            late_rate = pm.Exponential("late_rate", 1.0)
            rate = pm.math.switch(switchpoint >= years, early_rate, late_rate)
            with pytest.warns(Warning):
                disasters = pm.Poisson("disasters", rate, observed=disaster_data)

        return disaster_model, years

    def test_change_point_model(self, disaster_model):
        m, years = disaster_model

        ip = m.initial_point()
        ip["late_rate_log__"] += 1.0  # Make early and endpoint ip different

        ip.pop("switchpoint")
        ref_logp_fn = m.compile_logp(
            [m["switchpoint"], m["disasters_observed"], m["disasters_unobserved"]]
        )
        ref_logp = logsumexp([ref_logp_fn({**ip, **{"switchpoint": year}}) for year in years])

        marginal_m = marginalize(m, m["switchpoint"])

        with pytest.warns(UserWarning, match="There are multiple dependent variables"):
            marginal_m_logp = marginal_m.compile_logp(
                [marginal_m["disasters_observed"], marginal_m["disasters_unobserved"]]
            )(ip)
        np.testing.assert_almost_equal(marginal_m_logp, ref_logp)

    @pytest.mark.slow
    def test_change_point_model_sampling(self, disaster_model):
        m, _ = disaster_model

        rng = np.random.default_rng(211)

        with m:
            before_marg = pm.sample(chains=2, random_seed=rng).posterior.stack(
                sample=("draw", "chain")
            )

        marginal_m = marginalize(m, "switchpoint")

        with marginal_m:
            with pytest.warns(UserWarning, match="There are multiple dependent variables"):
                after_marg = pm.sample(chains=2, random_seed=rng).posterior.stack(
                    sample=("draw", "chain")
                )

        np.testing.assert_allclose(
            before_marg["early_rate"].mean(), after_marg["early_rate"].mean(), rtol=1e-2
        )
        np.testing.assert_allclose(
            before_marg["late_rate"].mean(), after_marg["late_rate"].mean(), rtol=1e-2
        )
        np.testing.assert_allclose(
            before_marg["disasters_unobserved"].mean(),
            after_marg["disasters_unobserved"].mean(),
            rtol=1e-2,
        )

    @pytest.mark.parametrize("univariate", (True, False))
    def test_vector_univariate_mixture(self, univariate):
        with Model() as m:
            idx = pm.Bernoulli("idx", p=0.5, shape=(2,) if univariate else ())

            def dist(idx, size):
                return pm.math.switch(
                    pm.math.eq(idx, 0),
                    pm.Normal.dist([-10, -10], 1),
                    pm.Normal.dist([10, 10], 1),
                )

            pm.CustomDist("norm", idx, dist=dist)

        marginal_m = marginalize(m, idx)
        logp_fn = marginal_m.compile_logp()

        if univariate:
            with pm.Model() as ref_m:
                pm.NormalMixture("norm", w=[0.5, 0.5], mu=[[-10, 10], [-10, 10]], shape=(2,))
        else:
            with pm.Model() as ref_m:
                pm.Mixture(
                    "norm",
                    w=[0.5, 0.5],
                    comp_dists=[
                        pm.MvNormal.dist([-10, -10], np.eye(2)),
                        pm.MvNormal.dist([10, 10], np.eye(2)),
                    ],
                    shape=(2,),
                )
        ref_logp_fn = ref_m.compile_logp()

        for test_value in (
            [-10, -10],
            [10, 10],
            [-10, 10],
            [-10, 10],
        ):
            pt = {"norm": test_value}
            np.testing.assert_allclose(logp_fn(pt), ref_logp_fn(pt))

    def test_k_censored_clusters_model(self):
        data = np.array([[-1.0, -1.0], [0.0, 0.0], [1.0, 1.0]])
        nobs = data.shape[0]
        n_clusters = 5

        def build_model(build_batched: bool) -> Model:
            coords = {
                "cluster": range(n_clusters),
                "ndim": ("x", "y"),
                "obs": range(nobs),
            }
            with Model(coords=coords) as m:
                if build_batched:
                    idx = pm.Categorical("idx", p=np.ones(n_clusters) / n_clusters, dims=["obs"])
                else:
                    idx = pm.math.stack(
                        [
                            pm.Categorical(f"idx_{i}", p=np.ones(n_clusters) / n_clusters)
                            for i in range(nobs)
                        ]
                    )

                mu_x = pm.Normal(
                    "mu_x",
                    dims=["cluster"],
                    transform=ordered,
                )
                mu_y = pm.Normal("mu_y", dims=["cluster"])
                mu = pm.math.stack([mu_x, mu_y], axis=-1)  # (cluster, ndim)
                mu_indexed = mu[idx, :]

                sigma = pm.HalfNormal("sigma")

                y = pm.Censored(
                    "y",
                    dist=pm.Normal.dist(mu_indexed, sigma),
                    lower=-3,
                    upper=3,
                    observed=data,
                    dims=["obs", "ndim"],
                )

            return m

        m = marginalize(build_model(build_batched=True), "idx")
        m.set_initval(m["mu_x"], np.linspace(-1, 1, n_clusters))

        ref_m = marginalize(build_model(build_batched=False), [f"idx_{i}" for i in range(nobs)])
        test_point = m.initial_point()
        np.testing.assert_almost_equal(
            m.compile_logp()(test_point),
            ref_m.compile_logp()(test_point),
        )


def test_unmarginalize():
    with pm.Model() as m:
        idx = pm.Bernoulli("idx", p=0.5)
        sub_idx = pm.Bernoulli("sub_idx", p=pt.as_tensor([0.3, 0.7])[idx])
        x = pm.Normal("x", mu=(idx + sub_idx) - 1)

    marginal_m = marginalize(m, [idx, sub_idx])
    assert not equivalent_models(marginal_m, m)

    unmarginal_m = unmarginalize(marginal_m)
    assert equivalent_models(unmarginal_m, m)

    unmarginal_idx_explicit = unmarginalize(marginal_m, ("idx", "sub_idx"))
    assert equivalent_models(unmarginal_idx_explicit, m)

    # Test partial unmarginalize
    unmarginal_idx = unmarginalize(marginal_m, "idx")
    assert equivalent_models(unmarginal_idx, marginalize(m, "sub_idx"))

    unmarginal_sub_idx = unmarginalize(marginal_m, "sub_idx")
    assert equivalent_models(unmarginal_sub_idx, marginalize(m, "idx"))


class TestRecoverMarginals:
    @pytest.mark.parametrize("explicit_model", (True, False))
    def test_basic(self, explicit_model):
        with Model() as m:
            sigma = pm.HalfNormal("sigma")
            p = np.array([0.5, 0.2, 0.3])
            k = pm.Categorical("k", p=p)
            mu = np.array([-3.0, 0.0, 3.0])
            mu_ = pt.as_tensor_variable(mu)
            y = pm.Normal("y", mu=mu_[k], sigma=sigma)

        marginal_m = marginalize(m, [k])

        rng = np.random.default_rng(211)

        with marginal_m:
            prior = pm.sample_prior_predictive(
                draws=20,
                random_seed=rng,
                return_inferencedata=False,
            )
            idata = InferenceData(posterior=dict_to_dataset(prior))

        if explicit_model:
            idata = recover_marginals(idata, model=marginal_m, return_samples=True)
        else:
            with marginal_m:
                idata = recover_marginals(idata, return_samples=True)

        post = idata.posterior
        assert "k" in post
        assert "lp_k" in post
        assert post.k.shape == post.y.shape
        assert post.lp_k.shape == (*post.k.shape, len(p))

        def true_logp(y, sigma):
            y = y.repeat(len(p)).reshape(len(y), -1)
            sigma = sigma.repeat(len(p)).reshape(len(sigma), -1)
            return log_softmax(
                np.log(p)
                + norm.logpdf(y, loc=mu, scale=sigma)
                + halfnorm.logpdf(sigma)
                + np.log(sigma),
                axis=1,
            )

        np.testing.assert_almost_equal(
            true_logp(post.y.values.flatten(), post.sigma.values.flatten()),
            post.lp_k[0].values,
        )
        np.testing.assert_almost_equal(logsumexp(post.lp_k, axis=-1), 0)

    def test_coords(self):
        """Test if coords can be recovered with marginalized value had it originally"""
        with Model(coords={"year": [1990, 1991, 1992]}) as m:
            sigma = pm.HalfNormal("sigma")
            idx = pm.Bernoulli("idx", p=0.75, dims="year")
            x = pm.Normal("x", mu=idx, sigma=sigma, dims="year")

        marginal_m = marginalize(m, [idx])
        rng = np.random.default_rng(211)

        with marginal_m:
            prior = pm.sample_prior_predictive(
                draws=20,
                random_seed=rng,
                return_inferencedata=False,
            )
            idata = InferenceData(
                posterior=dict_to_dataset({k: np.expand_dims(prior[k], axis=0) for k in prior})
            )

        with marginal_m:
            idata = recover_marginals(idata, return_samples=True)
        post = idata.posterior
        assert post.idx.dims == ("chain", "draw", "year")
        assert post.lp_idx.dims == ("chain", "draw", "year", "lp_idx_dim")

    def test_batched(self):
        """Test that marginalization works for batched random variables"""
        with Model() as m:
            sigma = pm.HalfNormal("sigma")
            idx = pm.Bernoulli("idx", p=0.7, shape=(3, 2))
            y = pm.Normal("y", mu=idx.T, sigma=sigma, shape=(2, 3))

        marginal_m = marginalize(m, [idx])

        rng = np.random.default_rng(211)

        with marginal_m:
            prior = pm.sample_prior_predictive(
                draws=20,
                random_seed=rng,
                return_inferencedata=False,
            )
            idata = InferenceData(
                posterior=dict_to_dataset({k: np.expand_dims(prior[k], axis=0) for k in prior})
            )

            idata = recover_marginals(idata, return_samples=True)
        post = idata.posterior
        assert post["y"].shape == (1, 20, 2, 3)
        assert post["idx"].shape == (1, 20, 3, 2)
        assert post["lp_idx"].shape == (1, 20, 3, 2, 2)

    def test_nested(self):
        """Test that marginalization works when there are nested marginalized RVs"""

        with Model() as m:
            idx = pm.Bernoulli("idx", p=0.75)
            sub_idx = pm.Bernoulli("sub_idx", p=pt.switch(pt.eq(idx, 0), 0.15, 0.95))
            sub_dep = pm.Normal("y", mu=idx + sub_idx, sigma=1.0)

        marginal_m = marginalize(m, [idx, sub_idx])

        rng = np.random.default_rng(211)

        with marginal_m:
            prior = pm.sample_prior_predictive(
                draws=20,
                random_seed=rng,
                return_inferencedata=False,
            )
            idata = InferenceData(posterior=dict_to_dataset(prior))

            idata = recover_marginals(idata, return_samples=True)
        post = idata.posterior
        assert "idx" in post
        assert "lp_idx" in post
        assert post.idx.shape == post.y.shape
        assert post.lp_idx.shape == (*post.idx.shape, 2)
        assert "sub_idx" in post
        assert "lp_sub_idx" in post
        assert post.sub_idx.shape == post.y.shape
        assert post.lp_sub_idx.shape == (*post.sub_idx.shape, 2)

        def true_idx_logp(y):
            idx_0 = np.log(0.85 * 0.25 * norm.pdf(y, loc=0) + 0.15 * 0.25 * norm.pdf(y, loc=1))
            idx_1 = np.log(0.05 * 0.75 * norm.pdf(y, loc=1) + 0.95 * 0.75 * norm.pdf(y, loc=2))
            return log_softmax(np.stack([idx_0, idx_1]).T, axis=1)

        np.testing.assert_almost_equal(
            true_idx_logp(post.y.values.flatten()),
            post.lp_idx[0].values,
        )

        def true_sub_idx_logp(y):
            sub_idx_0 = np.log(0.85 * 0.25 * norm.pdf(y, loc=0) + 0.05 * 0.75 * norm.pdf(y, loc=1))
            sub_idx_1 = np.log(0.15 * 0.25 * norm.pdf(y, loc=1) + 0.95 * 0.75 * norm.pdf(y, loc=2))
            return log_softmax(np.stack([sub_idx_0, sub_idx_1]).T, axis=1)

        np.testing.assert_almost_equal(
            true_sub_idx_logp(post.y.values.flatten()),
            post.lp_sub_idx[0].values,
        )
        np.testing.assert_almost_equal(logsumexp(post.lp_idx, axis=-1), 0)
        np.testing.assert_almost_equal(logsumexp(post.lp_sub_idx, axis=-1), 0)


def test_forward_after_sampling():
    # Regression test where forward graph was modified during sampling
    # Specifically `model.initial_point`, but we test the whole pipeline

    with pm.Model() as m:
        p_outlier = pm.Beta("p_outlier", 1, 1)
        is_outlier = pm.Bernoulli("is_outlier", p=p_outlier, shape=(10,))
        sigma = pm.Exponential("sigma", 1, shape=(2,))
        pm.Normal("y_hat", mu=0, sigma=sigma[is_outlier])

    marginalized_mod = marginalize(m, [is_outlier])

    # Check that model.initial_point() does not modify the inner graph of the MarginalRV
    marginal_rv = marginalized_mod["y_hat"]
    inner_outputs_before = marginal_rv.owner.op.fgraph.clone().outputs
    marginalized_mod.initial_point()
    inner_outputs_after = marginal_rv.owner.op.fgraph.outputs
    assert equal_computations_up_to_root(inner_outputs_before, inner_outputs_after)

    assert pm.draw(marginalized_mod["y_hat"]).shape == (10,)

    with marginalized_mod:
        pm.sample(tune=1, chains=1, draws=1, compute_convergence_checks=False)

    assert pm.draw(marginalized_mod["y_hat"]).shape == (10,)


def test_numpyro_compat():
    pytest.importorskip("numpyro")

    with pm.Model() as m:
        p_outlier = pm.Beta("p_outlier", 1, 1)
        is_outlier = pm.Bernoulli("is_outlier", p=p_outlier, shape=(10,))
        sigma = pm.Exponential("sigma", 1, shape=(2,))
        pm.Normal("y_hat", mu=0, sigma=sigma[is_outlier])

    with marginalize(m, [is_outlier]):
        pm.sample(nuts_sampler="numpyro", chains=1, tune=1, draws=1)



import pytest

from pymc_extras.statespace.core.properties import (
    CoordInfo,
    Data,
    DataInfo,
    Parameter,
    ParameterInfo,
    Shock,
    ShockInfo,
    State,
    StateInfo,
)
from pymc_extras.statespace.utils.constants import (
    ALL_STATE_AUX_DIM,
    ALL_STATE_DIM,
    OBS_STATE_AUX_DIM,
    OBS_STATE_DIM,
    SHOCK_AUX_DIM,
    SHOCK_DIM,
)


def test_property_str_formats_fields():
    p = Parameter(name="alpha", shape=(2,), dims=("param",))
    s = str(p).splitlines()
    assert s == [
        "name: alpha",
        "shape: (2,)",
        "dims: ('param',)",
        "constraints: None",
    ]


def test_info_lookup_contains_and_missing_key():
    params = [
        Parameter("a", (1,), ("d",)),
        Parameter("b", (2,), ("d",)),
        Parameter("c", (3,), ("d",)),
    ]
    info = ParameterInfo(params)

    assert info.get("b").name == "b"
    assert info["a"].shape == (1,)
    assert "c" in info

    with pytest.raises(KeyError) as e:
        _ = info["missing"]
    assert "No name 'missing'" in str(e.value)


def test_data_info_needs_exogenous_and_str():
    data = [
        Data("price", (10,), ("time",), is_exogenous=False),
        Data("x", (10,), ("time",), is_exogenous=True),
    ]
    info = DataInfo(data)

    assert info.needs_exogenous_data is True
    s = str(info)
    assert "data: ['price', 'x']" in s
    assert "needs exogenous data: True" in s

    no_exog = DataInfo([Data("y", (10,), ("time",), is_exogenous=False)])
    assert no_exog.needs_exogenous_data is False


def test_coord_info_make_defaults_from_component_and_types():
    class DummyComponent:
        state_names = ("x1", "x2")
        observed_states = ("x2",)
        shock_names = ("eps1",)

    coord_info = CoordInfo.default_coords_from_model(DummyComponent())

    expected = [
        (ALL_STATE_DIM, ("x1", "x2")),
        (ALL_STATE_AUX_DIM, ("x1", "x2")),
        (OBS_STATE_DIM, ("x2",)),
        (OBS_STATE_AUX_DIM, ("x2",)),
        (SHOCK_DIM, ("eps1",)),
        (SHOCK_AUX_DIM, ("eps1",)),
    ]

    assert len(coord_info) == 6
    assert isinstance(coord_info.items, tuple)

    for dim, labels in expected:
        assert dim in coord_info
        assert coord_info[dim].labels == labels


def test_state_info_and_shockinfo_basic():
    states = (
        State("x1", observed=True, shared=False),
        State("x2", observed=False, shared=True),
    )

    state_info = StateInfo(states)
    assert state_info[("x1", True)].observed is True
    s = str(state_info)

    assert "states: ['x1', 'x2']" in s
    assert "observed: [True, False]" in s

    shocks = (Shock("s1"), Shock("s2"))
    shock_info = ShockInfo(shocks)

    assert "s1" in shock_info
    assert shock_info["s2"].name == "s2"


def test_info_is_iterable_and_unpackable():
    items = (Parameter("p1", (1,), ("d",)), Parameter("p2", (2,), ("d",)))
    info = ParameterInfo(items)

    names = info.names
    assert names == ("p1", "p2")

    a, b = info.items
    assert a.name == "p1" and b.name == "p2"


def test_info_add_method():
    a_param = Parameter(name="a", shape=(1,), dims=("dim",))
    param_info = ParameterInfo(parameters=(a_param,))

    b_param = Parameter(name="b", shape=(1,), dims=("dim",))

    new_param_info = param_info.add(new_item=b_param)

    assert new_param_info.names == ("a", "b")


def test_info_merge_method():
    a_param = Parameter(name="a", shape=(1,), dims=("dim",))
    a_param_info = ParameterInfo(parameters=(a_param,))

    b_param = Parameter(name="b", shape=(1,), dims=("dim",))
    b_param_info = ParameterInfo(parameters=(b_param,))

    new_param_info = a_param_info.merge(b_param_info)

    assert new_param_info.names == ("a", "b")


def test_state_info_compound_key_allows_same_name_different_observed():
    """StateInfo uses compound key (name, observed) so same name with different observed is allowed."""
    states = (
        State("data", observed=False),
        State("other", observed=False),
        State("data", observed=True),
    )
    state_info = StateInfo(states)

    assert len(state_info) == 3
    assert state_info.unobserved_state_names == ("data", "other")
    assert state_info.observed_state_names == ("data",)
    assert state_info.names == ("data", "other", "data")

    assert state_info[("data", False)].observed is False
    assert state_info[("data", True)].observed is True

    assert "data" in state_info
    assert "other" in state_info
    assert "missing" not in state_info


def test_state_info_rejects_true_duplicates():
    """StateInfo rejects states with same name AND same observed value."""
    with pytest.raises(ValueError, match="Duplicate"):
        StateInfo(
            (
                State("x", observed=False),
                State("x", observed=False),
            )
        )


def test_state_info_merge_with_compound_keys():
    """Merging StateInfo works with compound keys."""
    info1 = StateInfo((State("a", observed=False), State("b", observed=True)))
    info2 = StateInfo((State("c", observed=False), State("a", observed=True)))

    merged = info1.merge(info2)

    assert len(merged) == 4
    assert ("a", False) in merged._index
    assert ("a", True) in merged._index
    assert ("b", True) in merged._index
    assert ("c", False) in merged._index


def test_state_info_merge_overwrite_duplicates():
    """Merging with overwrite_duplicates=True skips items with same compound key."""
    info1 = StateInfo((State("a", observed=False, shared=False),))
    info2 = StateInfo((State("a", observed=False, shared=True),))

    merged = info1.merge(info2, overwrite_duplicates=True)

    assert len(merged) == 1
    # Original kept, duplicate skipped
    assert merged[("a", False)].shared is False



import numpy as np
import statsmodels.api as sm


class LocalLinearTrend(sm.tsa.statespace.MLEModel):
    def __init__(self, endog, **kwargs):
        # Model order
        k_states = k_posdef = 2

        # Initialize the statespace
        super().__init__(endog, k_states=k_states, k_posdef=k_posdef, **kwargs)

        # Initialize the matrices
        self.ssm["design"] = np.array([1, 0])
        self.ssm["transition"] = np.array([[1, 1], [0, 1]])
        self.ssm["selection"] = np.eye(k_states)

        # Cache some indices
        self._state_cov_idx = ("state_cov", *np.diag_indices(k_posdef))

    @property
    def param_names(self):
        return ["sigma2.measurement", "sigma2.level", "sigma2.trend"]

    @property
    def start_params(self):
        return [np.std(self.endog)] * 3

    def transform_params(self, unconstrained):
        return unconstrained**2

    def untransform_params(self, constrained):
        return constrained**0.5

    def update(self, params, *args, **kwargs):
        params = super().update(params, *args, **kwargs)

        # Observation covariance
        self.ssm["obs_cov", 0, 0] = params[0]

        # State covariance
        self.ssm[self._state_cov_idx] = params[1:]



import pymc as pm
import pytensor
import pytensor.tensor as pt

from pymc import intX
from pymc.distributions.dist_math import check_parameters
from pymc.distributions.distribution import Continuous, SymbolicRandomVariable
from pymc.distributions.shape_utils import get_support_shape_1d
from pymc.logprob.abstract import _logprob
from pymc.pytensorf import normalize_rng_param
from pytensor.graph.basic import Node
from pytensor.tensor.random import multivariate_normal

floatX = pytensor.config.floatX
COV_ZERO_TOL = 0

lgss_shape_message = (
    "The LinearGaussianStateSpace distribution needs shape information to be constructed. "
    "Ensure that all input matrices have shape information specified."
)


def make_signature(sequence_names):
    states = "s"
    obs = "p"
    exog = "r"
    time = "t"
    state_and_obs = "n"

    matrix_to_shape = {
        "x0": (states,),
        "P0": (states, states),
        "c": (states,),
        "d": (obs,),
        "T": (states, states),
        "Z": (obs, states),
        "R": (states, exog),
        "H": (obs, obs),
        "Q": (exog, exog),
    }

    for matrix in sequence_names:
        base_shape = matrix_to_shape[matrix]
        matrix_to_shape[matrix] = (time, *base_shape)

    signature = ",".join(["(" + ",".join(shapes) + ")" for shapes in matrix_to_shape.values()])

    return f"{signature},[rng]->[rng],({time},{state_and_obs})"


class LinearGaussianStateSpaceRV(SymbolicRandomVariable):
    default_output = 1
    _print_name = ("LinearGuassianStateSpace", "\\operatorname{LinearGuassianStateSpace}")

    def update(self, node: Node):
        return {node.inputs[-1]: node.outputs[0]}


class _LinearGaussianStateSpace(Continuous):
    def __new__(
        cls,
        name,
        a0,
        P0,
        c,
        d,
        T,
        Z,
        R,
        H,
        Q,
        steps=None,
        sequence_names=None,
        append_x0=True,
        method="svd",
        **kwargs,
    ):
        # Ignore dims in support shape because they are just passed along to the "observed" and "latent" distributions
        # created by LinearGaussianStateSpace. This "combined" distribution shouldn't ever be directly used.
        steps = get_support_shape_1d(
            support_shape=steps,
            shape=None,
            dims=None,
            observed=kwargs.get("observed", None),
            support_shape_offset=0,
        )

        return super().__new__(
            cls,
            name,
            a0,
            P0,
            c,
            d,
            T,
            Z,
            R,
            H,
            Q,
            steps=steps,
            sequence_names=sequence_names,
            append_x0=append_x0,
            method=method,
            **kwargs,
        )

    @classmethod
    def dist(
        cls,
        a0,
        P0,
        c,
        d,
        T,
        Z,
        R,
        H,
        Q,
        steps=None,
        sequence_names=None,
        append_x0=True,
        method="svd",
        **kwargs,
    ):
        steps = get_support_shape_1d(
            support_shape=steps, shape=kwargs.get("shape", None), support_shape_offset=0
        )

        if steps is None:
            raise ValueError("Must specify steps or shape parameter")

        steps = pt.as_tensor_variable(intX(steps), ndim=0)

        return super().dist(
            [a0, P0, c, d, T, Z, R, H, Q, steps],
            sequence_names=sequence_names,
            append_x0=append_x0,
            method=method,
            **kwargs,
        )

    @classmethod
    def rv_op(
        cls,
        a0,
        P0,
        c,
        d,
        T,
        Z,
        R,
        H,
        Q,
        steps,
        size=None,
        rng=None,
        sequence_names=None,
        append_x0=True,
        method="svd",
    ):
        if sequence_names is None:
            sequence_names = []

        a0_, P0_, c_, d_, T_, Z_, R_, H_, Q_ = map(
            lambda x: x.type(), [a0, P0, c, d, T, Z, R, H, Q]
        )

        c_.name = "c"
        d_.name = "d"
        T_.name = "T"
        Z_.name = "Z"
        R_.name = "R"
        H_.name = "H"
        Q_.name = "Q"

        sequences = [
            x
            for x, name in zip([c_, d_, T_, Z_, R_, H_, Q_], ["c", "d", "T", "Z", "R", "H", "Q"])
            if name in sequence_names
        ]
        non_sequences = [x for x in [c_, d_, T_, Z_, R_, H_, Q_] if x not in sequences]

        rng = normalize_rng_param(rng)

        def sort_args(args):
            sorted_args = []

            # Inside the scan, outputs_info variables get a time step appended to their name
            # e.g. x -> x[t]. Remove this so we can identify variables by name.
            arg_names = [x.name.replace("[t]", "") for x in args]

            # c, d ,T, Z, R, H, Q is the "canonical" ordering
            for name in ["c", "d", "T", "Z", "R", "H", "Q"]:
                idx = arg_names.index(name)
                sorted_args.append(args[idx])

            return sorted_args

        n_seq = len(sequence_names)

        def step_fn(*args):
            seqs, (rng, state, *non_seqs) = args[:n_seq], args[n_seq:]

            c, d, T, Z, R, H, Q = sort_args((*seqs, *non_seqs))
            k = T.shape[0]
            a = state[:k]

            middle_rng, a_innovation = pm.MvNormal.dist(
                mu=0, cov=Q, rng=rng, method=method
            ).owner.outputs
            next_rng, y_innovation = pm.MvNormal.dist(
                mu=0, cov=H, rng=middle_rng, method=method
            ).owner.outputs

            a_mu = c + T @ a
            a_next = a_mu + R @ a_innovation

            y_mu = d + Z @ a_next
            y_next = y_mu + y_innovation

            next_state = pt.concatenate([a_next, y_next], axis=0)

            return next_rng, next_state

        Z_init = Z_ if Z_ in non_sequences else Z_[0]
        H_init = H_ if H_ in non_sequences else H_[0]

        init_x_ = pm.MvNormal.dist(a0_, P0_, rng=rng, method=method)
        init_y_ = pm.MvNormal.dist(Z_init @ init_x_, H_init, rng=rng, method=method)

        init_dist_ = pt.concatenate([init_x_, init_y_], axis=0)

        ss_rng, statespace = pytensor.scan(
            step_fn,
            outputs_info=[rng, init_dist_],
            sequences=None if len(sequences) == 0 else sequences,
            non_sequences=[*non_sequences],
            n_steps=steps,
            strict=True,
            return_updates=False,
        )

        if append_x0:
            statespace_ = pt.concatenate([init_dist_[None], statespace], axis=0)
            statespace_ = pt.specify_shape(statespace_, (steps + 1, None))
        else:
            statespace_ = statespace
            statespace_ = pt.specify_shape(statespace_, (steps, None))

        linear_gaussian_ss_op = LinearGaussianStateSpaceRV(
            inputs=[a0_, P0_, c_, d_, T_, Z_, R_, H_, Q_, steps, rng],
            outputs=[ss_rng, statespace_],
            extended_signature=make_signature(sequence_names),
        )

        linear_gaussian_ss = linear_gaussian_ss_op(a0, P0, c, d, T, Z, R, H, Q, steps, rng)
        return linear_gaussian_ss


class LinearGaussianStateSpace(Continuous):
    """
    Linear Gaussian Statespace distribution

    """

    def __new__(
        cls,
        name,
        a0,
        P0,
        c,
        d,
        T,
        Z,
        R,
        H,
        Q,
        *,
        steps,
        k_endog=None,
        sequence_names=None,
        append_x0=True,
        method="svd",
        **kwargs,
    ):
        dims = kwargs.pop("dims", None)
        latent_dims = None
        obs_dims = None
        if dims is not None:
            if len(dims) != 3:
                ValueError(
                    "LinearGaussianStateSpace expects 3 dims: time, all_states, and observed_states"
                )
            time_dim, state_dim, obs_dim = dims
            latent_dims = [time_dim, state_dim]
            obs_dims = [time_dim, obs_dim]

        latent_obs_combined = _LinearGaussianStateSpace(
            f"{name}_combined",
            a0,
            P0,
            c,
            d,
            T,
            Z,
            R,
            H,
            Q,
            steps=steps,
            sequence_names=sequence_names,
            append_x0=append_x0,
            method=method,
            **kwargs,
        )
        latent_obs_combined = pt.specify_shape(latent_obs_combined, (steps + int(append_x0), None))
        if k_endog is None:
            k_endog = cls._get_k_endog(H)
        latent_slice = slice(None, -k_endog)
        obs_slice = slice(-k_endog, None)

        latent_states = latent_obs_combined[..., latent_slice]
        obs_states = latent_obs_combined[..., obs_slice]

        latent_states = pm.Deterministic(f"{name}_latent", latent_states, dims=latent_dims)
        obs_states = pm.Deterministic(f"{name}_observed", obs_states, dims=obs_dims)

        return latent_states, obs_states

    @classmethod
    def dist(cls, a0, P0, c, d, T, Z, R, H, Q, *, steps=None, **kwargs):
        latent_obs_combined = _LinearGaussianStateSpace.dist(
            a0, P0, c, d, T, Z, R, H, Q, steps=steps, **kwargs
        )
        k_states = T.type.shape[0]

        latent_states = latent_obs_combined[..., :k_states]
        obs_states = latent_obs_combined[..., k_states:]

        return latent_states, obs_states

    @classmethod
    def _get_k_states(cls, T):
        k_states = T.type.shape[0]
        if k_states is None:
            raise ValueError(lgss_shape_message)
        return k_states

    @classmethod
    def _get_k_endog(cls, H):
        k_endog = H.type.shape[0]
        if k_endog is None:
            raise ValueError(lgss_shape_message)

        return k_endog


class KalmanFilterRV(SymbolicRandomVariable):
    default_output = 1
    _print_name = ("KalmanFilter", "\\operatorname{KalmanFilter}")
    extended_signature = "(t,s),(t,s,s),(t),[rng]->[rng],(t,s)"

    def update(self, node: Node):
        return {node.inputs[-1]: node.outputs[0]}


class SequenceMvNormal(Continuous):
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    @classmethod
    def dist(cls, mus, covs, logp, method="svd", **kwargs):
        mus, covs, logp = map(pt.as_tensor_variable, (mus, covs, logp))
        return super().dist([mus, covs, logp], method=method, **kwargs)

    @classmethod
    def rv_op(cls, mus, covs, logp, method="svd", size=None, rng=None):
        rng = normalize_rng_param(rng)
        logp_ = logp.type()

        mus_, covs_ = mus.type(), covs.type()
        seq_mvn_rng, mvn_seq = multivariate_normal(
            mean=mus_, cov=covs_, rng=rng, method=method
        ).owner.outputs

        mvn_seq_op = KalmanFilterRV(
            inputs=[mus_, covs_, logp_, rng], outputs=[seq_mvn_rng, mvn_seq], ndim_supp=2
        )

        mvn_seq = mvn_seq_op(mus, covs, logp, rng)

        return mvn_seq


@_logprob.register(KalmanFilterRV)
def sequence_mvnormal_logp(op, values, mus, covs, logp, rng, **kwargs):
    return check_parameters(
        logp,
        pt.eq(values[0].shape[0], mus.shape[0]),
        pt.eq(covs.shape[0], mus.shape[0]),
        msg="Observed data and parameters must have the same number of timesteps (dimension 0)",
    )



import warnings

from pymc_extras.statespace.models.structural.components.autoregressive import (
    Autoregressive,
)
from pymc_extras.statespace.models.structural.components.cycle import Cycle
from pymc_extras.statespace.models.structural.components.level_trend import LevelTrend
from pymc_extras.statespace.models.structural.components.measurement_error import MeasurementError
from pymc_extras.statespace.models.structural.components.regression import Regression
from pymc_extras.statespace.models.structural.components.seasonality import (
    FrequencySeasonality,
    TimeSeasonality,
)

_DEPRECATED_NAMES = {
    "LevelTrendComponent": LevelTrend,
    "CycleComponent": Cycle,
    "RegressionComponent": Regression,
    "AutoregressiveComponent": Autoregressive,
}


def __getattr__(name: str):
    if name in _DEPRECATED_NAMES:
        warnings.warn(
            f"{name} is deprecated and will be removed in a future release. "
            f"Use {_DEPRECATED_NAMES[name].__name__} instead.",
            FutureWarning,
            stacklevel=2,
        )
        return _DEPRECATED_NAMES[name]
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    "Autoregressive",
    "Cycle",
    "FrequencySeasonality",
    "LevelTrend",
    "MeasurementError",
    "Regression",
    "TimeSeasonality",
]



import warnings

import numpy as np
import pytensor.tensor as pt

from pymc_extras.statespace.core.properties import (
    Coord,
    Parameter,
    Shock,
    State,
)
from pymc_extras.statespace.models.structural.core import Component
from pymc_extras.statespace.models.structural.utils import order_to_mask


class Autoregressive(Component):
    r"""
    Autoregressive timeseries component

    Parameters
    ----------
    order: int or sequence of int

        If int, the number of lags to include in the model.
        If a sequence, an array-like of zeros and ones indicating which lags to include in the model.

    name: str, default "auto_regressive"
        A name for this autoregressive component. Used to label dimensions and coordinates.

    observed_state_names: list[str] | None, default None
        List of strings for observed state labels. If None, defaults to ["data"].

    share_states: bool, default False
        Whether latent states are shared across the observed states. If True, there will be only one set of latent
        states, which are observed by all observed states. If False, each observed state has its own set of
        latent states. This argument has no effect if `k_endog` is 1.

    Notes
    -----
    An autoregressive component can be thought of as a way o introducing serially correlated errors into the model.
    The process is modeled:

    .. math::
        x_t = \sum_{i=1}^p \rho_i x_{t-i}

    Where ``p``, the number of autoregressive terms to model, is the order of the process. By default, all lags up to
    ``p`` are included in the model. To disable lags, pass a list of zeros and ones to the ``order`` argumnet. For
    example, ``order=[1, 1, 0, 1]`` would become:

    .. math::
        x_t = \rho_1 x_{t-1} + \rho_2 x_{t-1} + \rho_4 x_{t-1}

    The coefficient :math:`\rho_3` has been constrained to zero.

    .. warning:: This class is meant to be used as a component in a structural time series model. For modeling of
              stationary processes with ARIMA, use ``statespace.BayesianSARIMAX``.

    Examples
    --------
    Model a timeseries as an AR(2) process with non-zero mean:

    .. code:: python

        from pymc_extras.statespace import structural as st
        import pymc as pm
        import pytensor.tensor as pt

        trend = st.LevelTrend(order=1, innovations_order=0)
        ar = st.Autoregressive(2)
        ss_mod = (trend + ar).build()

        with pm.Model(coords=ss_mod.coords) as model:
            P0 = pm.Deterministic('P0', pt.eye(ss_mod.k_states) * 10, dims=ss_mod.param_dims['P0'])
            intitial_trend = pm.Normal('initial_trend', sigma=10, dims=ss_mod.param_dims['initial_trend'])
            ar_params = pm.Normal('ar_params', dims=ss_mod.param_dims['ar_params'])
            sigma_ar = pm.Exponential('sigma_ar', 1, dims=ss_mod.param_dims['sigma_ar'])

            ss_mod.build_statespace_graph(data)
            idata = pm.sample(nuts_sampler='numpyro')

    """

    def __init__(
        self,
        order: int = 1,
        name: str = "auto_regressive",
        observed_state_names: list[str] | None = None,
        share_states: bool = False,
    ):
        if observed_state_names is None:
            observed_state_names = ["data"]

        k_endog = len(observed_state_names)
        k_endog_effective = k_posdef = 1 if share_states else k_endog

        order = order_to_mask(order)
        ar_lags = np.flatnonzero(order).ravel().astype(int) + 1
        k_states = len(order)

        self.share_states = share_states
        self.order = order
        self.ar_lags = ar_lags

        state_names = [f"L{i + 1}_{name}" for i in range(k_states)]

        super().__init__(
            name=name,
            k_endog=k_endog,
            k_states=k_states * k_endog_effective,
            k_posdef=k_posdef,
            measurement_error=True,
            combine_hidden_states=True,
            base_observed_state_names=observed_state_names,
            base_state_names=state_names,
            obs_state_idxs=np.tile(np.r_[[1.0], np.zeros(k_states - 1)], k_endog_effective),
            share_states=share_states,
        )

    def set_states(self) -> State | tuple[State, ...] | None:
        base_names = self.base_state_names
        observed_state_names = self.base_observed_state_names

        if self.share_states:
            state_names = [f"{name}[shared]" for name in base_names]
        else:
            state_names = [
                f"{name}[{state_name}]"
                for state_name in observed_state_names
                for name in base_names
            ]

        hidden_states = [State(name=name, observed=False, shared=True) for name in state_names]
        observed_states = [
            State(name=name, observed=True, shared=False) for name in observed_state_names
        ]
        return *hidden_states, *observed_states

    def set_parameters(self) -> Parameter | tuple[Parameter, ...] | None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        k_states = self.k_states // k_endog_effective  # this is also the number of AR lags

        ar_param = Parameter(
            name=f"params_{self.name}",
            shape=(k_endog_effective, k_states) if k_endog_effective > 1 else (k_states,),
            dims=(f"lag_{self.name}",)
            if k_endog_effective == 1
            else (
                f"endog_{self.name}",
                f"lag_{self.name}",
            ),
            constraints=None,
        )

        sigma_param = Parameter(
            name=f"sigma_{self.name}",
            shape=(k_endog_effective,) if k_endog_effective > 1 else (),
            dims=(f"endog_{self.name}",) if k_endog_effective > 1 else None,
            constraints="Positive",
        )

        return ar_param, sigma_param

    def set_shocks(self) -> Shock | tuple[Shock, ...] | None:
        observed_state_names = self.observed_state_names

        if self.share_states:
            shock_names = [f"{self.name}[shared]"]
        else:
            shock_names = [f"{self.name}[{obs_name}]" for obs_name in observed_state_names]

        return tuple(Shock(name=name) for name in shock_names)

    def set_coords(self) -> Coord | tuple[Coord, ...] | None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog
        observed_state_names = self.observed_state_names

        lag_coord = Coord(dimension=f"lag_{self.name}", labels=self.ar_lags.tolist())

        coord_container = [lag_coord]

        if k_endog_effective > 1:
            endog_coord = Coord(dimension=f"endog_{self.name}", labels=observed_state_names)
            coord_container.append(endog_coord)

        return tuple(coord_container)

    def make_symbolic_graph(self) -> None:
        k_endog = self.k_endog
        k_endog_effective = 1 if self.share_states else k_endog

        k_states = self.k_states // k_endog_effective
        k_posdef = self.k_posdef

        k_nonzero = int(sum(self.order))
        ar_params = self.make_and_register_variable(
            f"params_{self.name}",
            shape=(k_nonzero,) if k_endog_effective == 1 else (k_endog_effective, k_nonzero),
        )
        sigma_ar = self.make_and_register_variable(
            f"sigma_{self.name}", shape=() if k_endog_effective == 1 else (k_endog_effective,)
        )

        if k_endog_effective == 1:
            T = pt.eye(k_states, k=-1)
            ar_idx = (np.zeros(k_nonzero, dtype="int"), np.nonzero(self.order)[0])
            T = T[ar_idx].set(ar_params)

        else:
            transition_matrices = []

            for i in range(k_endog_effective):
                T = pt.eye(k_states, k=-1)
                ar_idx = (np.zeros(k_nonzero, dtype="int"), np.nonzero(self.order)[0])
                T = T[ar_idx].set(ar_params[i])
                transition_matrices.append(T)
            T = pt.specify_shape(
                pt.linalg.block_diag(*transition_matrices), (self.k_states, self.k_states)
            )

        self.ssm["transition", :, :] = T

        R = np.eye(k_states)
        R_mask = np.full((k_states,), False)
        R_mask[0] = True
        R = R[:, R_mask]

        self.ssm["selection", :, :] = pt.specify_shape(
            pt.linalg.block_diag(*[R for _ in range(k_endog_effective)]), (self.k_states, k_posdef)
        )

        Zs = [pt.zeros((1, k_states))[0, 0].set(1.0) for _ in range(k_endog)]

        if self.share_states:
            Z = pt.join(0, *Zs)
        else:
            Z = pt.linalg.block_diag(*Zs)
        self.ssm["design", :, :] = pt.specify_shape(Z, (k_endog, self.k_states))

        cov_idx = ("state_cov", *np.diag_indices(k_posdef))
        self.ssm[cov_idx] = sigma_ar**2


def __getattr__(name: str):
    if name == "AutoregressiveComponent":
        warnings.warn(
            "AutoregressiveComponent is deprecated and will be removed in a future release. "
            "Use Autoregressive instead.",
            FutureWarning,
            stacklevel=2,
        )
        return Autoregressive
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# ruff: noqa: I001

from pymc_extras.statespace.core.representation import PytensorRepresentation
from pymc_extras.statespace.core.statespace import PyMCStateSpace
from pymc_extras.statespace.core.compile import compile_statespace

__all__ = ["PyMCStateSpace", "PytensorRepresentation", "compile_statespace"]
